"""Gauges workflows for Wflow plugin."""

import logging
from typing import Optional

import geopandas as gpd
import numpy as np
import xarray as xr
from hydromt.gis import flw

logger = logging.getLogger(f"hydromt.{__name__}")


__all__ = ["gauge_map_uparea"]


def gauge_map_uparea(
    ds: xr.Dataset,
    gdf: gpd.GeoDataFrame,
    uparea_name: Optional[str] = "meta_upstream_area",
    mask: Optional[np.ndarray] = None,
    wdw: Optional[int] = 1,
    rel_error: float = 0.05,
    abs_error: float = 50,
    fillna: bool = False,
):
    """
    Snap point locations to grid cell.

    With smallest difference in upstream area within `wdw` around the
    original location if the local cell does not meet the error criteria.

    Both the upstream area variable named ``uparea_name`` in ``ds`` and ``gdf`` as
    well as ``abs_error`` should have the same unit (typically km2).

    Parameters
    ----------
    ds : xr.Dataset
        Dataset with upstream area variable.
    gdf : gpd.GeoDataFrame
        GeoDataFrame with gauge points and uparea column.
    uparea_name : str, optional
        Name of the upstream area variable in ``ds``, by default "meta_upstream_area".
    mask : np.ndarray, optional
        Mask cells to apply the uparea snapping, by default None.
    wdw : int, optional
        Window size around the original location to search for the best matching cell,
        by default 1.
    rel_error : float, optional
        Relative error threshold to accept the best matching cell, by default 0.05.
    abs_error : float, optional
        Absolute error threshold to accept the best matching cell, by default (50 km2).
    fillna : bool, optional
        Fill NaN values in gdf["uparea"] with uparea from ds, by default False.

    Returns
    -------
    da : xr.DataArray
        Gauge map with gauge points snapped to the best matching cell.
    idxs_out : np.ndarray
        Array of indices of the best matching cell.
    ids_out : np.ndarray
        Array of gauge point ids.

    """
    if uparea_name not in ds:
        raise ValueError(f"uparea_name {uparea_name} not found in ds.")
    if gdf["uparea"].isna().all():
        raise ValueError(
            "All gauges have NaN values for uparea. Use another method for snapping."
        )

    # Original number of gauges
    nb_gauges_before_snapping = gdf.index.size

    # Find if there are any nodata in gdf["uparea"]
    if gdf["uparea"].isna().any():
        # Index of gauges with NaN values
        nodata_gauges = list(gdf.index[gdf["uparea"].isna()])
        if fillna:
            logger.warning(
                f"Gauges with ID {nodata_gauges} have NaN values for uparea."
                "Uparea from wflow will be used and they won't be snapped."
            )
            # Replace nan values in gdf with uparea from wflow
            uparea_wflow = ds[uparea_name].raster.sample(gdf, wdw=0)
            gdf["uparea"] = gdf["uparea"].fillna(uparea_wflow.to_pandas())
        else:
            logger.warning(
                f"Gauges with ID {nodata_gauges} have NaN values for uparea."
                "They will be ignored."
            )
            gdf = gdf[gdf["uparea"].notna()]

    ds = ds.copy()
    # Add mask to ds
    if mask is not None:
        ds["mask"] = xr.DataArray(mask, dims=(ds.raster.y_dim, ds.raster.x_dim))

    ds_wdw = ds.raster.sample(gdf, wdw=wdw)
    # Mask valid cells in ds_wdw
    if mask is not None:
        ds_wdw[uparea_name] = ds_wdw[uparea_name].where(
            ds_wdw["mask"], ds_wdw[uparea_name].raster.nodata
        )
    logger.debug(
        f"Snapping gauges points to best matching uparea cell within wdw (size={wdw})."
    )
    upa0 = xr.DataArray(gdf["uparea"], dims=("index"))
    upa_dff = np.abs(ds_wdw[uparea_name].where(ds_wdw[uparea_name] > 0).load() - upa0)
    upa_check = np.logical_and((upa_dff / upa0) <= rel_error, upa_dff <= abs_error)

    # drop index will all nan values
    upa_dff = upa_dff.dropna(dim="index", how="all")
    upa_check = upa_check.sel(index=upa_dff.index)
    ds_wdw = ds_wdw.sel(index=upa_dff.index)

    # find best matching uparea cell in window
    i_wdw = upa_dff.argmin("wdw").load()
    idx_valid = np.nonzero(upa_check.isel(wdw=i_wdw).values)[0]
    if idx_valid.size < gdf.index.size:
        logger.warning(
            f"{idx_valid.size}/{gdf.index.size} gauge points successfully snapped."
        )
    i_wdw = i_wdw.isel(index=idx_valid)
    ds_out = ds_wdw.isel(wdw=i_wdw.load(), index=idx_valid)

    idxs_out = ds.raster.xy_to_idx(
        xs=ds_out[ds.raster.x_dim].values, ys=ds_out[ds.raster.y_dim].values
    )
    ids_out = ds_out.index.values

    # Derive gauge map
    da, idxs_out, ids_out = flw.gauge_map(
        ds,
        idxs=idxs_out,
        ids=ids_out,
        stream=None,
        flwdir=None,
    )

    # Final message
    logger.info(
        f"Snapped {idxs_out.size}/{nb_gauges_before_snapping} gauge points "
        "to best matching uparea cell."
    )

    return da, idxs_out, ids_out



"""Wflow netcdf_grid output component."""

import logging
from pathlib import Path

import xarray as xr
from hydromt.model import Model
from hydromt.model.components import GridComponent

from hydromt_wflow.components import utils

__all__ = ["WflowOutputGridComponent"]

logger = logging.getLogger(f"hydromt.{__name__}")


class WflowOutputGridComponent(GridComponent):
    """ModelComponent class for Wflow netcdf_grid output.

    This class is used for reading the Wflow netcdf_grid output.

    The overall output netcdf_grid component data stored in the ``data`` property of
    this class is of the hydromt.gis.raster.RasterDataset type which is an extension of
    xarray.Dataset for regular grid.
    """

    def __init__(
        self,
        model: Model,
        *,
        filename: str = "output_grid.nc",
        region_component: str | None = None,
    ):
        """
        Initialize the Wflow netcdf_grid output component.

        Parameters
        ----------
        model : Model
            HydroMT model instance.
        filename : str, optional
            Default path relative to the root where the netcdf_grid output file will be
            read and written. By default 'output_grid.nc'.
        region_component : str, optional
            Name of the region component to align the states data with. If provided,
            the states data will be aligned with the grid of this component.
            By default None.
        """
        super().__init__(
            model=model,
            filename=filename,
            region_component=region_component,
            region_filename=None,
        )

    ## I/O methods
    def read(self):
        """Read netcdf_grid model output at root/dir_output/filename.

        Checks the path of the file in the config toml using both
        ``output.netcdf_grid.path`` and ``dir_output``. If not found uses the default
        path ``output_grid.nc`` in the root folder.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments to be passed to the `read_nc` method.
        """
        # Sort which path/ filename is actually the one used
        # Hierarchy is: 1: config, 2: default from component
        p = self.model.config.get_value("output.netcdf_grid.path") or self._filename
        # Check for input dir
        p_input = Path(self.model.config.get_value("dir_output", fallback=""), p)
        # Add root
        p_input = Path(self.root.path, p_input)

        if not p_input.is_file():
            logger.warning(
                f"Netcdf grid output file {p_input} not found, skip reading."
            )
            return

        logger.info(f"Read netcdf_grid output from {p_input}")
        with xr.open_dataset(p_input, chunks={"time": 30}, decode_coords="all") as ds:
            self.set(ds)

    def write(self):
        """
        Skip writing output files.

        Output files are model results and are therefore not written by HydroMT.
        """
        logger.warning("netcdf_grid is an output of Wflow and will not be written.")

    ## Set methods
    def set(
        self,
        data: xr.DataArray | xr.Dataset,
        name: str | None = None,
    ):
        """Add data to netcdf_grid output.

        All layers in data must have identical spatial coordinates to existing
        staticmaps.

        Parameters
        ----------
        data: xarray.DataArray or xarray.Dataset
            new map layer to add to netcdf_grid output
        name: str, optional
            Name of new map layer, this is used to overwrite the name of a DataArray
            and ignored if data is a Dataset
        """
        # Check the type
        if not isinstance(data, (xr.DataArray, xr.Dataset)):
            raise TypeError("Data must be an xarray Dataset or DataArray")

        # Check that time dimension is present
        if "time" not in data.dims:
            raise ValueError("'time' dimension not found in data")

        if self._region_component is not None:
            # Ensure the data is aligned with the region component (staticmaps)
            data = utils.align_grid_with_region(
                data,
                region_grid=self._get_grid_data(),
                region_component_name=self._region_component,
            )

        # Call set of parent class
        super().set(
            data=data,
            name=name,
        )



"""Workflow for water demand."""

import logging
from typing import Optional

import geopandas as gpd
import numpy as np
import xarray as xr
from hydromt.gis import raster_utils
from hydromt.model.processes.grid import grid_from_constant, grid_from_geodataframe

logger = logging.getLogger(f"hydromt.{__name__}")

__all__ = [
    "allocation_areas",
    "domestic",
    "domestic_from_population",
    "other_demand",
    "surfacewaterfrac_used",
    "irrigation",
    "irrigation_from_vector",
]

map_vars = {
    "dom": "domestic",
    "ind": "industry",
    "lsk": "livestock",
}


def create_grid_from_bbox(
    bbox: list[float],
    res: float,
    crs: int,
    align: True,
    y_name: str = "y",
    x_name: str = "x",
):
    """Create grid from bounding box and target resolution."""
    xmin, ymin, xmax, ymax = bbox
    res = abs(res)
    if align:
        xmin = round(xmin / res) * res
        ymin = round(ymin / res) * res
        xmax = round(xmax / res) * res
        ymax = round(ymax / res) * res
    xcoords = np.linspace(
        xmin + res / 2,
        xmax - res / 2,
        num=round((xmax - xmin) / res),
        endpoint=True,
    )
    ycoords = np.flip(
        np.linspace(
            ymin + res / 2,
            ymax - res / 2,
            num=round((ymax - ymin) / res),
            endpoint=True,
        )
    )
    coords = {y_name: ycoords, x_name: xcoords}
    grid = raster_utils.full(
        coords=coords,
        nodata=np.nan,
        dtype=np.float32,
        name="data",
        attrs={},
        crs=crs,
        lazy=True,
    )

    return grid


def domestic(
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    popu: Optional[xr.Dataset] = None,
    original_res: Optional[float] = None,
) -> tuple:
    """Create domestic water demand maps.

    Parameters
    ----------
    ds : xr.Dataset
        Raw domestic data dataset.
    ds_like : xr.Dataset
        Dataset at wflow model domain and resolution.
    popu : xr.Dataset
        Population dataset (number of people per gridcell).
    original_res : float
        Original resolution of ds. If provided, ds will be upscaled before
        downsampling with population.

    Returns
    -------
    tuple
        Domestic data at model resolution, Population data at model resolution.
    """
    # Mask no data values
    ds = ds.raster.mask_nodata()
    # Convert population to density and reproject to model
    if popu is not None:
        popu = popu.raster.mask_nodata().fillna(0)
        popu_density = popu / popu.raster.area_grid()
        popu_density.name = "population_density"
        popu_scaled = popu_density.raster.reproject_like(
            ds_like,
            method="average",
        )
        # Back to cap per cell
        popu_scaled = popu_scaled * popu_scaled.raster.area_grid()
        popu_scaled.name = "population"
    else:
        popu_scaled = None

    # Reproject to model resolution
    # Simple reprojection
    ds_scaled = ds.raster.reproject_like(
        ds_like,
        method="average",
    )
    # Downscale with population if available
    if popu is not None:
        # Reproject ds to original resolution if needed before downscaling
        if original_res is not None:
            # Create empty grid at the original resolution
            dda = create_grid_from_bbox(
                bbox=ds.raster.bounds,
                res=original_res,
                crs=ds_like.raster.crs,
                align=True,
                y_name=ds.raster.y_dim,
                x_name=ds.raster.x_dim,
            )
            # Use (weighted) average to scale demands in mm!
            ds = ds.raster.reproject_like(
                dda,
                method="average",
            )
        # Check if population is of higher res than ds (else no need to downscale)
        if abs(popu.raster.res[0]) < abs(ds.raster.res[0]):
            # Convert ds from mm to m3
            ds_m3 = ds * ds.raster.area_grid() * 1000
            # Get the number of capita per cell
            popu_ds = popu.raster.reproject_like(
                ds_m3,
                method="sum",
            )
            # Get m3 per capita
            ds_m3_per_cap = ds_m3 / popu_ds
            ds_m3_per_cap = ds_m3_per_cap.where(popu_ds > 0, 0)
            # Downscale to model resolution
            ds_m3_per_cap_model = ds_m3_per_cap.raster.reproject_like(
                ds_like,
                method="average",
            )
            # Get m3 per cell
            ds_m3_model = ds_m3_per_cap_model * popu_scaled
            # Get back to mm
            ds_scaled = ds_m3_model / ds_m3_model.raster.area_grid() / 1000
            # Reset the nodata attribute
            for var in ds_scaled.data_vars:
                ds_scaled[var].raster.set_nodata(ds[var].raster.nodata)

    return ds_scaled, popu_scaled


def domestic_from_population(
    popu: xr.Dataset,
    ds_like: xr.Dataset,
    gross_per_capita: float | list[float],
    net_per_capita: float | list[float],
) -> tuple:
    """Create domestic water demand maps from statitics per capita.

    Parameters
    ----------
    popu : xr.Dataset
        Population dataset (number of people per gridcell).
    ds_like : xr.Dataset
        Dataset at wflow model domain and resolution.
    gross_per_capita : float, list[float]
        Gross domestic water demand per capita in m3/day.
    net_per_capita : float, list[float], optional
        Net domestic water demand per capita in m3/day.

    Returns
    -------
    tuple
        Domestic data at model resolution, Population data at model resolution.
    """
    # Convert population to density and reproject to model
    popu = popu.raster.mask_nodata().fillna(0)
    popu_density = popu / popu.raster.area_grid()
    popu_density.name = "population_density"
    popu_scaled = popu_density.raster.reproject_like(
        ds_like,
        method="average",
    )
    # Back to cap per cell
    popu_scaled = popu_scaled * popu_scaled.raster.area_grid()
    popu_scaled.name = "population"

    # Create the demand maps
    gross_per_capita = np.atleast_1d(gross_per_capita)
    net_per_capita = np.atleast_1d(net_per_capita)
    if len(gross_per_capita) != len(net_per_capita):
        raise ValueError("Gross and net per capita demands must have the same length.")

    # Non-cyclic
    if len(gross_per_capita) == 1:
        ds_demand = popu_scaled * gross_per_capita[0]
        ds_demand.name = "domestic_gross"
        ds_demand = ds_demand.to_dataset()
        ds_demand["domestic_net"] = popu_scaled * net_per_capita[0]
    # Cyclic
    else:
        ds_demand = xr.concat(
            [popu_scaled * gross_per_capita[i] for i in range(len(gross_per_capita))],
            dim="time",
        )
        ds_demand["time"] = [i for i in range(len(gross_per_capita))]
        ds_demand = ds_demand.assign_coords(
            time=[i for i in range(len(gross_per_capita))]
        )
        ds_demand = ds_demand.to_dataset(name="domestic_gross")
        ds_demand["domestic_net"] = xr.concat(
            [popu_scaled * net_per_capita[i] for i in range(len(net_per_capita))],
            dim="time",
        )

    # Convert from m3/day to mm/day
    for var in ds_demand.data_vars:
        ds_demand[var] = ds_demand[var] / ds_demand.raster.area_grid() * 1000

    return ds_demand, popu_scaled


def other_demand(
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    ds_method: str = "average",
) -> xr.Dataset:
    """Create non-irrigation water demand maps.

    Parameters
    ----------
    ds : xr.Dataset
        Raw demand data dataset.
    ds_like : xr.Dataset
        Dataset at wflow model domain and resolution.
    ds_method : str
        Demand data resampling method. Default is 'average'.

    Returns
    -------
    xr.Dataset
        Demand data at model resolution.
    """
    # Reproject to up or downscale
    non_irri_scaled = ds.raster.reproject_like(
        ds_like,
        method=ds_method,
    )

    # Mask data to fill nodata gaps with 0
    non_irri_scaled = non_irri_scaled.raster.mask_nodata()
    for var in non_irri_scaled.data_vars:
        non_irri_scaled[var] = non_irri_scaled[var].where(
            ~np.isnan(non_irri_scaled[var]), 0
        )

    return non_irri_scaled


def allocation_areas(
    ds_like: xr.Dataset,
    waterareas: gpd.GeoDataFrame,
    basins: gpd.GeoDataFrame,
    priority_basins: bool = True,
    minimum_area: float = 50.0,
) -> tuple[xr.DataArray, gpd.GeoDataFrame]:
    """Create water allocation area.

    Based on current wflow model domain and resolution and making use of
    the model basins and administrative boundaries.

    Parameters
    ----------
    ds_like : xr.DataArray
        A grid covering the wflow model domain and the rivers.

        * Required variables: ['rivmsk', 'basins']
    waterareas : gpd.GeoDataFrame
        Administrative boundaries, e.g. sovereign nations.
    basins : gpd.GeoDataFrame
        The wflow model basins.
    priority_basins : bool
        For basins that do not contain river cells after intersection with the admin
        boundaries, the priority_basins flag can be used to decide if these basins
        should be merged with the closest downstream basin (True, default) or with any
        large enough basin in the same administrative area (False).
    minimum_area : float
        Minimum area of the subbasins to keep in km2. Default is 50 km2.

    Returns
    -------
    xr.DataArray
        The water demand allocation areas.
    gpd.GeoDataFrame
        The water demand allocation areas as geodataframe.
    """
    # Intersect the basins with the admin boundaries
    subbasins = basins.overlay(waterareas, how="intersection")
    # Create a unique index
    subbasins.index = np.arange(1, len(subbasins) + 1)
    subbasins.index.name = "uid"

    # After intersection, some cells at the coast are not included in the subbasins
    # convert to raster to fill na with nearest value
    subbasins_raster = ds_like.raster.rasterize(subbasins, col_name="uid", nodata=0)
    subbasins_raster = subbasins_raster.raster.interpolate_na(
        method="nearest", extrapolate=True
    )
    # Mask the wflow subcatch after extrapolation
    subbasins_raster = subbasins_raster.where(
        ds_like["basins"] != ds_like["basins"].raster.nodata,
        subbasins_raster.raster.nodata,
    )

    # Convert back to geodataframe and prepare the new unique index for the subbasins
    subbasins = subbasins_raster.raster.vectorize()
    if priority_basins:
        # Redefine the uid, equivalent to exploding the geometries
        subbasins.index = np.arange(1, len(subbasins) + 1)
    else:
        # Keep the uid of the intersection with the admin bounds
        subbasins.index = subbasins["value"].astype(int)
    subbasins.index.name = "uid"

    # Rasterize the subbasins
    da_subbasins = ds_like.raster.rasterize(subbasins, col_name="uid", nodata=0)
    # Mask with river cells
    da_subbasins_to_keep = np.unique(
        da_subbasins.raster.mask_nodata().where(ds_like["rivmsk"] > 0, np.nan).values
    )
    # Remove the nodata value from the list
    da_subbasins_to_keep = np.int32(
        da_subbasins_to_keep[~np.isnan(da_subbasins_to_keep)]
    )
    # Filter with a minimum area for the subbasins to keep
    subbasins_to_keep = subbasins[subbasins.index.isin(da_subbasins_to_keep)]
    if subbasins_to_keep.crs.is_geographic:
        subbasins_to_keep = subbasins_to_keep.to_crs(3857)
    # Get the area of the subbasins
    subbasins_to_keep["area"] = subbasins_to_keep.area
    # Filter the subbasins based on the minimum area
    subbasins_to_keep = subbasins_to_keep[
        subbasins_to_keep["area"] > (minimum_area * 1e6)
    ]
    # Get the unique index of the subbasins to keep
    da_subbasins_to_keep = subbasins_to_keep.index.values

    # Create the water allocation map starting with the subbasins that contain a river
    allocation_areas = da_subbasins.where(
        da_subbasins.isin(da_subbasins_to_keep), da_subbasins.raster.nodata
    )
    # Use nearest to fill the nodata values for subbasins without rivers
    allocation_areas = allocation_areas.raster.interpolate_na(
        method="nearest", extrapolate=True
    )
    allocation_areas = allocation_areas.where(
        ds_like["basins"] != ds_like["basins"].raster.nodata,
        allocation_areas.raster.nodata,
    )
    allocation_areas.name = "allocation_areas"

    # Create the equivalent geodataframe
    allocation_areas_gdf = allocation_areas.raster.vectorize()
    allocation_areas_gdf.index = allocation_areas_gdf["value"].astype(int)
    allocation_areas_gdf.index.name = "uid"

    return allocation_areas, allocation_areas_gdf


def surfacewaterfrac_used(
    gwfrac_raw: xr.DataArray,
    da_like: xr.DataArray,
    waterareas: xr.DataArray,
    gwbodies: Optional[xr.DataArray] = None,
    ncfrac: Optional[xr.DataArray] = None,
    interpolate: bool = False,
    mask_and_scale_gwfrac: bool = True,
) -> xr.DataArray:
    """Create surface water fraction map.

    Parameters
    ----------
    gwfrac_raw : xr.DataArray
        Raw groundwater fraction data map.
    da_like : xr.DataArray
        Wflow model like grid.
    waterareas : xr.DataArray
        Water source areas.
    gwbodies : xr.DataArray, optional
        Groundwater bodies map. Either 0 or 1. If None, assumes 1.
    ncfrac : xr.DataArray, optional
        Non-conventional water fraction. If None, assumes 0.
    interpolate : bool
        Interpolate missing data values within wflow model domain.
    mask_and_scale_gwfrac : bool, optional
        If True, gwfrac will be masked for areas with no groundwater bodies. To keep
        the average gwfrac used over waterareas similar after the masking, gwfrac
        for areas with groundwater bodies can increase. If False, gwfrac will be
        used as is. By default True.

    Returns
    -------
    xr.DataArray
        Surface water fraction.
    """
    # Resample the data to model resolution
    gwfrac_raw_mr = gwfrac_raw.raster.reproject_like(
        da_like,
        method="average",
    )
    # Only reproject waterareas if needed
    if not waterareas.raster.identical_grid(da_like):
        waterareas_mr = waterareas.raster.reproject_like(
            da_like,
            method="mode",
        )
    else:
        waterareas_mr = waterareas
    # Set nodata values to zeros
    waterareas_mr = waterareas_mr.where(
        waterareas_mr != waterareas_mr.raster.nodata,
        0,
    )

    if gwbodies is not None:
        gwbodies_mr = gwbodies.raster.reproject_like(
            da_like,
            method="mode",
        )
    else:
        gwbodies_mr = grid_from_constant(
            grid_like=da_like,
            constant=1,
            name="gwbodies",
            nodata=-9999,
            dtype=np.int32,
        )
    if ncfrac is not None:
        ncfrac_mr = ncfrac.raster.reproject_like(
            da_like,
            method="average",
        )
    else:
        ncfrac_mr = grid_from_constant(
            grid_like=da_like,
            constant=0,
            name="ncfrac",
            nodata=-9999,
            dtype=np.float32,
        )

    # Get the fractions based on area count
    x, y = np.nonzero(~np.isnan(gwfrac_raw_mr))
    if mask_and_scale_gwfrac:
        gw_pixels = np.take(
            np.bincount(
                waterareas_mr.values[x, y],
                weights=gwbodies_mr.values[x, y],
            ),
            waterareas_mr.values[x, y],
        )
        a_pixels = np.take(
            np.bincount(
                waterareas_mr.values[x, y],
                weights=gwbodies_mr.values[x, y] * 0.0 + 1.0,
            ),
            waterareas_mr.values[x, y],
        )
        scale_factor = a_pixels / (gw_pixels + 0.01)
    else:
        scale_factor = 1.0

    # Determine the groundwater fraction
    gwfrac_val = np.minimum(
        gwfrac_raw_mr.values[x, y] * scale_factor,
        1 - ncfrac_mr.values[x, y],
    )
    if mask_and_scale_gwfrac:
        gwfrac_val[np.nonzero(gwbodies_mr.values[x, y] == 0)] = 0
    # invert to get surface water frac
    swfrac_val = np.maximum(np.minimum(1 - gwfrac_val - ncfrac_mr.values[x, y], 1), 0)

    # create the dataarray for the fraction
    swfrac = xr.full_like(
        gwfrac_raw_mr,
        fill_value=np.nan,
        dtype=np.float32,
    ).load()
    swfrac.name = "demand_surface_water_ratio"
    swfrac.attrs = {"_FillValue": -9999}
    swfrac.raster.set_crs(da_like.raster.crs)

    # Set and interpolate the values
    swfrac = swfrac.copy()  # to avoid read-only error
    swfrac.values[x, y] = swfrac_val
    if interpolate:
        swfrac = swfrac.raster.interpolate_na(method="linear", extrapolate=True)
    else:
        # Fill na with 1 (no groundwater or nc sources used)
        swfrac = swfrac.fillna(1.0)

    # Set the nodata values based on the dem of the model (da_like)
    swfrac = swfrac.where(da_like != da_like.raster.nodata, -9999)

    # Return surface water frac
    return swfrac


def classify_pixels(
    da_irr: xr.DataArray,
    da_crop_model: xr.DataArray,
    threshold: float,
    nodata_value: float | int = -9999,
):
    """Classifies pixels based on a (fractional) threshold.

    Pixels with a value above this threshold are set to be irrigated, and pixels below
    this value as rainfed.

    Parameters
    ----------
    da_irr: xr.DataArray
        Data with irrigation mask
    da_crop_model: xr.DataArray
        Layer of the masked cropland in wflow model
    threshold: float
        Threshold above which pixels are classified as irrigated
    nodata_value: float, int = -9999
        Value to be used for nodata

    Returns
    -------
    irr_map: xr.DataArray
        Mask with classified pixels
    """
    # Check if the resolution of the irrigation map is higher than the model
    if abs(da_irr.raster.res[0]) < abs(da_crop_model.raster.res[0]):
        # Use area to resample
        # Convert irr map to an area map
        da_area = da_irr * da_irr.raster.area_grid()
        # Resample to model grid and sum areas
        da_area2model = da_area.raster.reproject_like(da_crop_model, method="sum")

        # Calculate relative area
        relative_area = da_area2model / da_area2model.raster.area_grid()

        # Classify pixels with a 1 where it exceeds the threshold, and 0 where it doesnt
        da_irr_model = xr.where(relative_area >= threshold, 1, 0)

    # Else resample using nearest
    else:
        da_irr_model = da_irr.raster.reproject_like(da_crop_model, method="nearest")

    # Find cropland pixels that are irrigated
    irr_map = xr.where(
        np.logical_and(da_irr_model == 1, da_crop_model != da_crop_model.raster.nodata),
        1,
        0,
    )
    # Fix nodata values (will be used as catchment mask)
    irr_map.raster.set_nodata(nodata_value)

    return irr_map


def calc_lai_threshold(da_lai, threshold, dtype=np.int32, na_value=-9999):
    """
    Calculate irrigation trigger based on LAI threshold.

    Trigger is set to 1 when the LAI is bigger than 20% of the variation (set by the
    threshold value).

    Parameters
    ----------
    da_lai
        Dataarray with LAI values
    threshold
        Value to be used as threshold
    dtype
        Datatype for the resulting map
    na_value
        Value for nodata

    Returns
    -------
    trigger:
        Maps with a value of 1 where the LAI indicates growing season, and 0 for all
        other pixels
    """
    # Compute min and max of LAI
    lai_min = da_lai.min(dim="time")
    lai_max = da_lai.max(dim="time")
    # Determine critical threshold
    ct = lai_min + threshold * (lai_max - lai_min)
    # Set missing value and dtype
    trigger = xr.where(da_lai >= ct, 1, 0)
    trigger = trigger.where(~ct.isnull())
    trigger = trigger.fillna(na_value).astype(dtype)
    trigger.raster.set_nodata(np.dtype(dtype).type(na_value))

    return trigger


def irrigation(
    da_irrigation: xr.DataArray,
    ds_like: xr.Dataset,
    irrigation_value: list[int],
    cropland_class: list[int],
    paddy_class: list[int] = [],
    area_threshold: float = 0.6,
    lai_threshold: float = 0.2,
):
    """
    Prepare irrigation maps for paddy and non paddy.

    Parameters
    ----------
    da_irrigation: xr.DataArray
        Irrigation map
    ds_like: xr.Dataset
        Dataset at wflow model domain and resolution.

        * Required variables: ['landuse', 'LAI']
    irrigation_value: list[int]
        Values that indicate irrigation in da_irrigation.
    cropland_class: list[int]
        Values that indicate cropland in landuse map.
    paddy_class: list[int]
        Values that indicate paddy fields in landuse map.
    area_threshold: float
        Threshold for the area of a pixel to be classified as irrigated (fraction of
        the cell covered).
    lai_threshold: float
        Threshold for the LAI value to be classified as growing season.

    Returns
    -------
    ds_irrigation: xr.Dataset
        Dataset with paddy and non-paddy irrigation maps: [
        'demand_paddy_irrigated_mask',
        'demand_nonpaddy_irrigated_mask', 'demand_paddy_irrigation_trigger',
        'demand_nonpaddy_irrigation_trigger']
    """
    # Check that the landuse map is available
    if "landuse" not in ds_like:
        raise ValueError("Landuse map is required in ds_like.")
    landuse = ds_like["landuse"].copy()

    # Create the irrigated areas mask
    irrigation_mask = da_irrigation.isin(irrigation_value).astype(float)

    # Get cropland mask from landuse
    logger.info("Preparing irrigated areas map for non paddy.")
    nonpaddy = landuse.where(landuse.isin(cropland_class), landuse.raster.nodata)
    nonpaddy_areas = classify_pixels(irrigation_mask, nonpaddy, area_threshold)
    ds_irrigation = nonpaddy_areas.to_dataset(name="demand_nonpaddy_irrigated_mask")

    # Get paddy mask from landuse
    if len(paddy_class) > 0:
        logger.info("Preparing irrigated areas map for paddy.")
        paddy = landuse.where(landuse.isin(paddy_class), landuse.raster.nodata)
        paddy_areas = classify_pixels(irrigation_mask, paddy, area_threshold)
        ds_irrigation["demand_paddy_irrigated_mask"] = paddy_areas

    # Calculate irrigation trigger based on LAI
    logger.info("Calculating irrigation trigger.")
    if "LAI" not in ds_like:
        raise ValueError("LAI map is required in ds_like.")
    lai = ds_like["LAI"].copy()
    trigger = calc_lai_threshold(lai, lai_threshold)

    # Mask trigger with paddy and nonpaddy
    ds_irrigation["demand_nonpaddy_irrigation_trigger"] = trigger.where(
        nonpaddy_areas == 1, 0
    )
    if len(paddy_class) > 0:
        ds_irrigation["demand_paddy_irrigation_trigger"] = trigger.where(
            paddy_areas == 1, 0
        )

    return ds_irrigation


def irrigation_from_vector(
    gdf_irrigation: gpd.GeoDataFrame,
    ds_like: xr.Dataset,
    cropland_class: list[int],
    paddy_class: list[int] = [],
    area_threshold: float = 0.6,
    lai_threshold: float = 0.2,
):
    """
    Prepare irrigation maps for paddy and non paddy from geodataframe.

    Parameters
    ----------
    gdf_irrigation: gpd.GeoDataFrame
        Shapefile with irrigation areas.
    ds_like: xr.Dataset
        Dataset at wflow model domain and resolution.

        * Required variables: ['landuse', 'LAI']
    irrigation_value: list[int]
        Values that indicate irrigation in gdf_irrigation.
    cropland_class: list[int]
        Values that indicate cropland in landuse map.
    paddy_class: list[int]
        Values that indicate paddy fields in landuse map.
    area_threshold: float
        Threshold for the area of a pixel to be classified as irrigated (fraction of
        the cell covered).
    lai_threshold: float
        Threshold for the LAI value to be classified as growing season.

    Returns
    -------
    ds_irrigation: xr.Dataset
        Dataset with paddy and non-paddy irrigation maps: [
        'demand_paddy_irrigated_mask',
        'demand_nonpaddy_irrigated_mask', 'demand_paddy_irrigation_trigger',
        'demand_nonpaddy_irrigation_trigger']

    See Also
    --------
    irrigation
    """
    # Rasterize the irrigation geometries
    ds_irrigation = grid_from_geodataframe(
        grid_like=ds_like,
        gdf=gdf_irrigation,
        rasterize_method="fraction",
        rename="irrigated_area",
        nodata=-1,
        mask_name=None,
    )

    # Only keep fraction that are above area_threshold and convert to 1
    da_irrigation = ds_irrigation["irrigated_area"]
    da_irrigation = da_irrigation.where(
        da_irrigation >= area_threshold, da_irrigation.raster.nodata
    )
    da_irrigation = da_irrigation.where(da_irrigation == da_irrigation.raster.nodata, 1)

    # Call the raster method
    ds_irrigation = irrigation(
        da_irrigation=da_irrigation,
        ds_like=ds_like,
        irrigation_value=[1],
        cropland_class=cropland_class,
        paddy_class=paddy_class,
        area_threshold=area_threshold,
        lai_threshold=lai_threshold,
    )

    return ds_irrigation



"""Wflow netcdf_scalar output component."""

import logging
from pathlib import Path
from typing import cast

import xarray as xr
from hydromt.model import Model
from hydromt.model.components import ModelComponent

__all__ = ["WflowOutputScalarComponent"]

logger = logging.getLogger(f"hydromt.{__name__}")


class WflowOutputScalarComponent(ModelComponent):
    """ModelComponent class for Wflow netcdf_scalar output.

    This class is used for reading the Wflow netcdf_scalar output.

    The overall output netcdf_scalar component data stored in the ``data`` property of
    this class is a xarray.Dataset.
    """

    def __init__(
        self,
        model: Model,
        *,
        filename: str = "output_scalar.nc",
    ):
        """
        Initialize the Wflow netcdf_scalar output component.

        Parameters
        ----------
        model : Model
            HydroMT model instance.
        filename : str, optional
            Default path relative to the root where the netcdf_scalar output file will
            be read and written. By default 'output_scalar.nc'.
        """
        self._data: xr.Dataset | None = None
        self._filename: str = filename

        super().__init__(
            model=model,
        )

    ## I/O methods
    def read(self):
        """Read netcdf_scalar model output at root/dir_output/filename.

        Checks the path of the file in the config toml using both
        ``output.netcdf_scalar.path`` and ``dir_output``. If not found uses the default
        path ``output_scalar.nc`` in the root folder.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments to be passed to the `read_nc` method.
        """
        # Sort which path/ filename is actually the one used
        # Hierarchy is: 1: config, 2: default from component
        p = self.model.config.get_value("output.netcdf_scalar.path") or self._filename
        # Check for input dir
        p_input = Path(self.model.config.get_value("dir_output", fallback=""), p)
        # Add root
        p_input = Path(self.root.path, p_input)

        if not p_input.is_file():
            logger.warning(
                f"Netcdf scalar output file {p_input} not found, skip reading."
            )
            return

        logger.info(f"Read netcdf_scalar output from {p_input}")
        with xr.open_dataset(p_input, chunks={"time": 30}) as ds:
            self.set(ds)

    def write(self):
        """
        Skip writing output files.

        Output files are model results and are therefore not written by HydroMT.
        """
        logger.warning("netcdf_scalar is an output of Wflow and will not be written.")

    ## Set methods
    def _initialize(self, skip_read: bool = False) -> None:
        """Initialize netcdf_scalar output object."""
        if self._data is None:
            self._data = xr.Dataset()
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    def set(
        self,
        data: xr.DataArray | xr.Dataset,
        name: str | None = None,
    ):
        """Add data to netcdf_scalar output.

        Parameters
        ----------
        data: xarray.DataArray or xarray.Dataset
            new map layer to add to netcdf_scalar output
        name: str, optional
            Name of new map layer, this is used to overwrite the name of a DataArray
            and ignored if data is a Dataset
        """
        self._initialize()
        assert self._data is not None

        if isinstance(data, xr.DataArray):
            # NOTE _name can be different from _data.name !
            if name is not None:
                data.name = name
            elif name is None and data.name is not None:
                name = data.name
            elif data.name is None and name is None:
                raise ValueError("Name required for DataArray.")
            data = data.to_dataset()

        if not isinstance(data, xr.Dataset):
            raise TypeError(f"cannot set data of type {type(data).__name__}")

        if len(self._data) == 0:  # empty grid
            self._data = data
        else:
            for dvar in data.data_vars:
                if dvar in self._data:
                    logger.warning(f"Replacing netcdf_scalar output: {dvar}")
                self._data[dvar] = data[dvar]

    ## Properties
    @property
    def data(self) -> xr.Dataset:
        """Model netcdf_scalar output data as xarray.Dataset."""
        if self._data is None:
            self._initialize()
        assert self._data is not None
        return self._data

    ## Testing
    def test_equal(self, other: ModelComponent) -> tuple[bool, dict[str, str]]:
        """Test if two components are equal.

        Parameters
        ----------
        other : ModelComponent
            The component to compare against.

        Returns
        -------
        Tuple[bool, Dict[str, str]]
            True if the components are equal, and a dict with the associated errors per
            property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_grid = cast(WflowOutputScalarComponent, other)
        try:
            xr.testing.assert_allclose(self.data, other_grid.data)
        except AssertionError as e:
            errors["data"] = str(e)

        return len(errors) == 0, errors



"""Wflow csv output component."""

import logging
from pathlib import Path

import xarray as xr
from hydromt.model import Model
from hydromt.model.components import DatasetsComponent
from hydromt.model.steps import hydromt_step

from hydromt_wflow import utils
from hydromt_wflow.components.staticmaps import WflowStaticmapsComponent

__all__ = ["WflowOutputCsvComponent"]

logger = logging.getLogger(f"hydromt.{__name__}")


class WflowOutputCsvComponent(DatasetsComponent):
    """ModelComponent class for Wflow csv output.

    This class is used for reading the Wflow csv output.

    The overall output csv component data stored in the ``data`` property of
    this class is of the hydromt.gis.vector.GeoDataset type which is an extension of
    xarray.Dataset for vector data.
    """

    def __init__(
        self,
        model: Model,
        *,
        filename: str = "output.csv",
        locations_component: str | None = None,
    ):
        """
        Initialize the Wflow csv output component.

        Parameters
        ----------
        model : Model
            HydroMT model instance.
        filename : str, optional
            Default path relative to the root where the csv output file will be
            read and written. By default 'output.csv'.
        locations_component : str, optional
            Name of the locations component to use for reading the output and mapping
            them to right geo-locations.
        """
        self._locations_component = locations_component

        super().__init__(
            model=model,
            filename=filename,
        )

    ## I/O methods
    @hydromt_step
    def read(self):
        """Read csv model output at root/dir_output/filename.

        Checks the path of the file in the config toml using both
        ``output.csv.path`` and ``dir_output``. If not found uses the default
        path ``output.csv`` in the root folder.
        """
        self.root._assert_read_mode()
        self._initialize(skip_read=True)

        # Sort which path/ filename is actually the one used
        # Hierarchy is: 1: config, 2: default from component
        p = self.model.config.get_value("output.csv.path") or self._filename
        # Check for input dir
        p_input = Path(self.model.config.get_value("dir_output", fallback=""), p)

        # Link with root
        csv_filename = Path(self.root.path, p_input)

        # Check if the file exist
        if not csv_filename.is_file():
            logger.warning(f"CSV output file {csv_filename} not found, skip reading.")
            return

        # Read
        staticmaps = self._get_locations_data()

        # Check if staticmaps data is not empty
        if len(staticmaps) == 0:
            logger.warning("Staticmaps data is empty, skip reading.")
            return

        csv_dict = utils.read_csv_output(
            csv_filename, config=self.model.config.data, maps=staticmaps
        )
        for key in csv_dict:
            # Add to data
            self.set(csv_dict[f"{key}"])

    def write(self):
        """
        Skip writing output files.

        Output files are model results and are therefore not written by HydroMT.
        """
        logger.warning("output_csv is an output of Wflow and will not be written.")

    ## Set methods

    ## Other
    def _get_locations_data(self) -> xr.Dataset:
        """Get locations data to geo-localize the csv outputs."""
        if self._locations_component is not None:
            reference_component = self.model.get_component(self._locations_component)
            if not isinstance(reference_component, WflowStaticmapsComponent):
                raise ValueError(
                    "Unable to find the referenced staticmaps component: "
                    f"'{self._locations_component}'."
                )
            if reference_component.data is None:
                raise ValueError(
                    "Unable to get grid from the referenced locations component: "
                    f"'{self._locations_component}'."
                )
            return reference_component.data



"""Rootzoneclim workflows for Wflow plugin."""

import logging
from typing import Optional

import numpy as np
import pandas as pd
import pyflwdir
import xarray as xr
from hydromt.gis import flw
from scipy import optimize

logger = logging.getLogger(f"hydromt.{__name__}")

__all__ = ["rootzoneclim"]


def determine_budyko_curve_terms(
    ds_sub_annual,
):
    """
    Determine the Budyko terms.

    (Note that the discharge coefficient and evaporative
    index for the future climate, if present, are not correct yet and will
    be adjusted at a later stage).

    Parameters
    ----------
    ds_sub_annual : xr.Dataset
        Dataset containing per subcatchment the annual precipitation,
        potential evaporation and specific discharge sums.

    Returns
    -------
    ds_sub_annual : xr.Dataset
        Similar to input, but containing the discharge coefficient, aridity
        index and the evaporative index as long term averages.

    """
    ds_sub_annual["discharge_coeff"] = (
        ds_sub_annual["specific_Q"] / ds_sub_annual["precip_mean"]
    ).mean("time", skipna=True)
    ds_sub_annual["aridity_index"] = (
        ds_sub_annual["pet_mean"] / ds_sub_annual["precip_mean"]
    ).mean("time", skipna=True)
    ds_sub_annual["evap_index"] = 1 - ds_sub_annual["discharge_coeff"]

    # Make sure Ea = P - Q < Ep. If not, we will not use that subcatchment for
    # the calculations.
    ds_sub_annual["discharge_coeff"] = ds_sub_annual["discharge_coeff"].where(
        ds_sub_annual["evap_index"] < ds_sub_annual["aridity_index"]
    )
    ds_sub_annual["aridity_index"] = ds_sub_annual["aridity_index"].where(
        ds_sub_annual["evap_index"] < ds_sub_annual["aridity_index"]
    )
    ds_sub_annual["evap_index"] = ds_sub_annual["evap_index"].where(
        ds_sub_annual["evap_index"] < ds_sub_annual["aridity_index"]
    )

    # Final check, if a coefficient < 0.0 or >>1.0, set to nan
    ds_sub_annual["discharge_coeff"] = (
        ds_sub_annual["discharge_coeff"]
        .where(ds_sub_annual["discharge_coeff"] < 10.0)
        .where(ds_sub_annual["discharge_coeff"] > 0.0)
    )
    ds_sub_annual["aridity_index"] = (
        ds_sub_annual["aridity_index"]
        .where(ds_sub_annual["aridity_index"] < 10.0)
        .where(ds_sub_annual["aridity_index"] > 0.0)
    )
    ds_sub_annual["evap_index"] = (
        ds_sub_annual["evap_index"]
        .where(ds_sub_annual["evap_index"] < 10.0)
        .where(ds_sub_annual["evap_index"] > 0.0)
    )

    return ds_sub_annual


def determine_omega(ds_sub_annual):
    """
    Determine the omega parameter.

    This function uses the Zhang function (as defined in Teuling et al., 2019)
    to determine the omega parameter for ds_sub_annual.

    Teuling, A. J., de Badts, E. A. G., Jansen, F. A., Fuchs, R., Buitink, J.,
    Hoek van Dijke, A. J., and Sterling, S. M.: Climate change, reforestation/
    afforestation, and urbanization impacts on evapotranspiration and streamflow
    in Europe, Hydrol. Earth Syst. Sci., 23, 3631–3652,
    https://doi.org/10.5194/hess-23-3631-2019, 2019.

    Parameters
    ----------
    ds_sub_annual : xr.Dataset
        Dataset containing at least the discharge coefficient, aridity
        index and evaporative index for the different forcing types.

    Returns
    -------
    ds_sub_annual : xr.Dataset
        Same as above, but with the omega parameter added. The omega parameter
        is the same for all forcing types and based on the observations.

    """
    # Construct the output variable in the dataset
    ds_sub_annual = ds_sub_annual.assign(
        omega=lambda ds_sub_annual: ds_sub_annual["discharge_coeff"] * np.nan
    )
    # Load the aridity index and evaporative index as np arrays (this saves
    # calculation time in the loop below)
    # calculate omega for "obs" and "cc_hist":
    if "cc_hist" in ds_sub_annual.forcing_type:
        forcing_types = ["obs", "cc_hist"]
    else:
        forcing_types = ["obs"]

    for forcing_type in forcing_types:
        aridity_index = ds_sub_annual.sel(forcing_type=forcing_type)[
            "aridity_index"
        ].values
        evap_index = ds_sub_annual.sel(forcing_type=forcing_type)["evap_index"].values

        # Set the temporary omega variable
        omega = np.zeros((len(ds_sub_annual.index),))

        for subcatch_index_nr in range(len(ds_sub_annual.index)):
            if evap_index[subcatch_index_nr] > 0:  # make sure evap index is not nan.
                try:
                    omega_temp = optimize.brentq(
                        Zhang,
                        1e-6,  # changed from 1e-12 to 1e-6 for numerical stability
                        100,
                        args=(
                            aridity_index[subcatch_index_nr],
                            evap_index[subcatch_index_nr],
                        ),
                    )
                    omega[subcatch_index_nr] = omega_temp
                # possible error that occurs: "ValueError: f(a) and f(b) must
                # have different signs") -- increase a and b range solves the issue.
                except ValueError:
                    logger.warning("No value for omega could be derived.")
                    omega[subcatch_index_nr] = np.nan
            else:
                omega[subcatch_index_nr] = np.nan

        # Add omega to the xr dataset
        ds_sub_annual["omega"].loc[dict(forcing_type=forcing_type)] = omega

    return ds_sub_annual


def determine_Peffective_Interception_explicit(ds_sub, Imax, intercep_vars_sub=None):
    """
    Determine the effective precipitation, interception evaporation and canopy storage.

    Based on (daily) values of precipitation and potential evaporation.

    Parameters
    ----------
    ds_sub : xr.Dataset
        Dataset containing precipitation and potential evaporation
        (precip_mean and pet_mean).
    Imax : float
        The maximum interception storage capacity [mm].
    intercep_vars_sub : xr.Dataarray
        Dataarray from staticmaps containing the Imax values for a number
        of time steps, e.g. every month. If present, this is used to determine
        Imax per time step. The default is None.

    Returns
    -------
    ds_sub : xr.Dataset
        same as above, but with effective precipitation, interception evaporation
        and canopy storage added.
    """
    # Make the output dataset ready for the output
    ds_sub["evap_interception"] = xr.full_like(
        ds_sub["precip_mean"], fill_value=ds_sub["precip_mean"].vector.nodata
    )
    ds_sub["evap_interception"].load()
    ds_sub["precip_effective"] = xr.full_like(
        ds_sub["precip_mean"],
        fill_value=ds_sub["precip_mean"].vector.nodata,
    )
    ds_sub["precip_effective"].load()
    ds_sub["canopy_storage"] = xr.full_like(
        ds_sub["precip_mean"],
        fill_value=ds_sub["precip_mean"].vector.nodata,
    )
    ds_sub["canopy_storage"].load()

    # Calculate it per forcing type
    for forcing_type in ds_sub["forcing_type"].values:
        nr_time_steps = len(
            ds_sub[["precip_mean", "pet_mean"]]
            .sel(forcing_type=forcing_type)
            .dropna("time")
            .time
        )

        # Add new empty variables that will be filled in the loop
        evap_interception = np.zeros(
            (len(ds_sub.sel(forcing_type=forcing_type)["index"]), nr_time_steps)
        )
        precip_effective = np.zeros(
            (len(ds_sub.sel(forcing_type=forcing_type)["index"]), nr_time_steps)
        )
        canopy_storage = np.zeros(
            (len(ds_sub.sel(forcing_type=forcing_type)["index"]), nr_time_steps)
        )

        # Load the potential evaporation and precipitation into memory to speed up
        # the subsequent for loop.
        # make sure order of coord is the same.
        Epdt = (
            ds_sub.sel(forcing_type=forcing_type)["pet_mean"]
            .dropna("time")
            .transpose("index", "time")
            .values
        )
        Pdt = (
            ds_sub.sel(forcing_type=forcing_type)["precip_mean"]
            .dropna("time")
            .transpose("index", "time")
            .values
        )
        # Loop through the time steps and determine the variables per time step.
        for i in range(0, nr_time_steps):
            if intercep_vars_sub is not None:
                # TODO: assumption LAI contains monthly data,
                # change this for future
                month = pd.to_datetime(ds_sub.time[i].values).month
                Imax = intercep_vars_sub["Imax"].sel(time=month)
            # Determine the variables with a simple interception reservoir approach
            canopy_storage[:, i] = canopy_storage[:, i] + Pdt[:, i]
            precip_effective[:, i] = np.maximum(0, canopy_storage[:, i] - Imax)
            canopy_storage[:, i] = canopy_storage[:, i] - precip_effective[:, i]
            evap_interception[:, i] = np.minimum(Epdt[:, i], canopy_storage[:, i])
            canopy_storage[:, i] = canopy_storage[:, i] - evap_interception[:, i]

            # Update Si for the next time step
            if i < nr_time_steps - 1:
                canopy_storage[:, i + 1] = canopy_storage[:, i]

        # insert in ds for the time that is available in each forcing type for
        # precip and pet
        time_forcing_type = (
            ds_sub[["precip_mean", "pet_mean"]]
            .sel(forcing_type=forcing_type)
            .dropna("time")
            .time
        )
        ds_sub["evap_interception"].loc[
            dict(forcing_type=forcing_type, time=time_forcing_type)
        ] = evap_interception
        ds_sub["precip_effective"].loc[
            dict(forcing_type=forcing_type, time=time_forcing_type)
        ] = precip_effective
        ds_sub["canopy_storage"].loc[
            dict(forcing_type=forcing_type, time=time_forcing_type)
        ] = canopy_storage

    return ds_sub


def determine_storage_deficit(ds_sub, correct_cc_deficit):
    """
    Determine the storage deficit for every time step.

    Also for the subcatchment location and dataset in ds_sub.

    Parameters
    ----------
    ds_sub : xr.Dataset
        Dataset containing the daily or higher-resolution data.

    Returns
    -------
    ds_sub : xr.Dataset
        Same as above, but containing the storage deficits per time step for all
        forcing types.
    """
    # make sure the order of the coordinates is always the same.
    # Calculate it per forcing type

    ds_sub["storage_deficit"] = xr.full_like(
        ds_sub["precip_mean"],
        fill_value=ds_sub["precip_mean"].vector.nodata,
    )
    ds_sub["storage_deficit"].load()

    for forcing_type in ds_sub["forcing_type"].values:
        time_forcing_type = (
            ds_sub["precip_effective"]
            .sel(forcing_type=forcing_type)
            .dropna("time")
            .time.values
        )
        transpiration = (
            ds_sub["transpiration"]
            .sel(forcing_type=forcing_type, time=time_forcing_type)
            .transpose("index", "time")
            .values
        )
        precip_effective = (
            ds_sub["precip_effective"]
            .sel(forcing_type=forcing_type, time=time_forcing_type)
            .transpose("index", "time")
            .values
        )

        storage_deficit = np.zeros((len(ds_sub["index"]), len(time_forcing_type)))

        # Determine the storage deficit per time step
        for i in range(1, len(time_forcing_type)):
            storage_deficit[:, i] = np.minimum(
                0,
                storage_deficit[:, i - 1]
                + (precip_effective[:, i] - transpiration[:, i]),
            )

        # Create the storage deficit data array
        ds_sub["storage_deficit"].loc[
            dict(forcing_type=forcing_type, time=time_forcing_type)
        ] = storage_deficit

    # If there are climate projections present, adjust the storage deficit for
    # the future projections based on Table S1 in Bouaziz et al., 2002, HESS.
    # cc_hist remains as is (see Table S1)
    if (len(ds_sub.forcing_type) > 1) & (correct_cc_deficit == True):
        if (
            len(
                ds_sub["precip_mean"]
                .sel(forcing_type=["cc_fut", "cc_hist"])
                .dropna("time")
                .time
            )
            > 0
        ):
            ds_sub["storage_deficit"].loc[dict(forcing_type="cc_fut")] = ds_sub[
                "storage_deficit"
            ].sel(forcing_type="obs") + np.minimum(
                0.0,
                ds_sub["storage_deficit"].sel(forcing_type="cc_fut")
                - ds_sub["storage_deficit"].sel(forcing_type="cc_hist"),
            )
        else:
            logger.warning(
                "Time period of cc_hist and cc_fut does not overlap. \
Correct_cc_deficit not applied."
            )

    return ds_sub


def fut_discharge_coeff(ds_sub_annual, correct_cc_deficit):
    """
    Determine the future discharge coefficient.

    Based on a given omega value (generally same as in the
    current-climate observations), the aridity index of the future climate and the
    difference in Q and PET between the simulated historical and
    simulated future climate.

    Parameters
    ----------
    ds_sub_annual : xr.Dataset
        Dataset containing at least the future omega and aridity index.

    Returns
    -------
    ds_sub_annual : xr.Dataset
        Similar to previous, but containing the new future discharge coefficient
        and evaporative index.
    """
    # Determine the delP and delEP for cc_fut
    Ep = ds_sub_annual["pet_mean"].sel(forcing_type="obs").mean("time")
    P = ds_sub_annual["precip_mean"].sel(forcing_type="obs").mean("time")
    delP = (
        ds_sub_annual["precip_mean"]
        .sel(forcing_type=["cc_hist", "cc_fut"])
        .mean("time")
        .diff("forcing_type")
        .sel(forcing_type="cc_fut")
    )
    delEp = (
        ds_sub_annual["pet_mean"]
        .sel(forcing_type=["cc_hist", "cc_fut"])
        .mean("time")
        .diff("forcing_type")
        .sel(forcing_type="cc_fut")
    )

    # Determine the difference in discharge between the observations and the
    # future climate simulations
    Q_obs_mean = ds_sub_annual["specific_Q"].mean("time")
    # if correct_cc_deficit=False -- omega is from cc_hist, else omega is from obs
    if correct_cc_deficit == True:
        omega = ds_sub_annual["omega"].sel(forcing_type="obs")
    else:
        omega = ds_sub_annual["omega"].sel(forcing_type="cc_hist")
    aridity_index_fut = (Ep + delEp) / (P + delP)
    change_Q_total = Zhang_future(omega, aridity_index_fut) * (P + delP) - Q_obs_mean

    # Determine the discharge coefficient for the future climate
    ds_sub_annual["discharge_coeff"].loc[dict(forcing_type="cc_fut")] = (
        Q_obs_mean + change_Q_total
    ) / (P + delP)

    return ds_sub_annual


def gumbel_su_calc_xr(
    storage_deficit_annual, storage_deficit_count, return_period, threshold
):
    """
    Determine the Gumbel distribution of the annual maximum storage deficits.

    For a set of return periods.

    Parameters
    ----------
    storage_deficit_annual : xr.Dataset
        Dataset with the minimum deficit per year as a positive value.
    storage_deficit_count : xr.Dataset
        Dataset containing the number of days with data per year. This
        indicates on how many days a minimum in year_min_storage_deficit is
        based.
    return_period : list
        List with one or more values indicating the return period(s) (in
        years) for which the rootzone storage depth should be calculated.
    threshold : int
        Required minimum number of days in a year containing data to take the
        year into account for the calculation..

    Returns
    -------
    gumbel : xr.Dataset
        Dataset, similar to year_min_storage_deficit, but containing the
        root-zone storage capacity per return period.

    """
    # Only take the years into account that contain more than [threshold] days
    # of data.
    storage_deficit_annual = storage_deficit_annual.where(
        storage_deficit_count > threshold
    )

    # Calculate the mean and standard deviation for the Gumbel distribution
    annual_mean = storage_deficit_annual.mean("time", skipna=True)
    annual_std = storage_deficit_annual.std("time", skipna=True)

    # Calculate alpha and mu
    alpha = (np.sqrt(6.0) * annual_std) / np.pi
    mu = annual_mean - 0.5772 * alpha

    # Create the output dataset
    gumbel = storage_deficit_annual.to_dataset(name="storage_deficit")

    # Set the return periods
    RP = return_period
    gumbel["RP"] = RP
    gumbel["rootzone_storage"] = (
        ("index", "forcing_type", "RP"),
        np.zeros(
            (len(gumbel["index"]), len(gumbel["forcing_type"]), len(gumbel["RP"]))
        ),
    )

    # Determine the root zone storage for different return periods using alpha
    # and mu
    gumbel["yt"] = -np.log(-np.log(1 - (1 / gumbel["RP"])))
    gumbel["rootzone_storage"] = mu + alpha * gumbel["yt"]

    return gumbel


def Zhang(omega, aridity_index, evap_index):
    """
    Calculate omega according to Zhang.

    This is the Zhang equation with omega as in Teuling et al., 2019.
    This function is used to get omega for historical situations when
    aridity_index and evap_index are known (assuming evap_index = 1 - discharge_coeff).
    This equation is solved for Zhang eq = 0.

    Parameters
    ----------
    omega : float
        Parameter of the Budyko curve.
    aridity_index : float
        The aridity index.
    evap_index : float
        The evaporative index.

    Returns
    -------
    Value 0.

    References
    ----------
    Fu, B.: On the calculation of the evaporation from land surface,
    Scientia Atmospherica Sinica, 5, 23–31, 1981 (in Chinese).

    Teuling, A. J., de Badts, E. A. G., Jansen, F. A., Fuchs, R., Buitink, J.,
    Hoek van Dijke, A. J., and Sterling, S. M.: Climate change, reforestation/
    afforestation, and urbanization impacts on evapotranspiration and streamflow
    in Europe, Hydrol. Earth Syst. Sci., 23, 3631–3652,
    https://doi.org/10.5194/hess-23-3631-2019, 2019.

    Zhang, L., Hickel, K., Dawes, W. R., Chiew, F. H., Western, A. W., and
    Briggs, P. R.: A rational function approach for estimating mean annual
    evapotranspiration, Water Resour. Res., 40, 1–14,
    https://doi.org/10.1029/2003WR002710, 2004.
    """
    return 1 + aridity_index - (1 + aridity_index**omega) ** (1 / omega) - evap_index


def Zhang_future(omega, aridity_index):
    """
    Calculate discharge coefficients according to Zhang.

    Once omega has been derived for historical situations, it can be used to
    derive the new discharge coefficient when the future aridity index (Ep/P)
    is known. This will allow the discharge coefficient to shift over the same
    line as the historical situation.

    Parameters
    ----------
    omega : xr.Datarray
        Datarray containing the future omega values per sub catchment.
    aridity_index : xr.Datarray
        Datarray containing the future aridity index values per sub
        catchment.

    Returns
    -------
    Datarray containing the future discharge coefficient.
    """
    return -(aridity_index - (1 + aridity_index**omega) ** (1 / omega))


def check_inputs(
    start_hydro_year, start_field_capacity, dsrun, ds_obs, ds_cc_hist, ds_cc_fut
):
    # Start with some initial checks
    list_of_months = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ]
    if start_hydro_year not in list_of_months:
        raise ValueError(
            f"start_hydro_year not in {list_of_months}: provide a valid month"
        )
    if start_field_capacity not in list_of_months:
        raise ValueError(
            f"start_field_capacity not in {list_of_months}: provide a valid month"
        )

    if "discharge" not in list(dsrun.keys()):
        raise ValueError("Variable discharge not in run_fn")

    if "precip" not in list(ds_obs.keys()):
        raise ValueError("Variable precip not in forcing_obs_fn")

    if "pet" not in list(ds_obs.keys()):
        raise ValueError("Variable pet not in forcing_obs_fn")
    if ds_cc_hist is not None:
        if "precip" not in list(ds_cc_hist.keys()):
            raise ValueError("Variable precip not in forcing_cc_hist_fn")

        if "pet" not in list(ds_cc_hist.keys()):
            raise ValueError("Variable pet not in forcing_cc_hist_fn")
    if ds_cc_fut is not None:
        if "precip" not in list(ds_cc_fut.keys()):
            raise ValueError("Variable precip not in forcing_cc_fut_fn")

        if "pet" not in list(ds_cc_fut.keys()):
            raise ValueError("Variable pet not in forcing_cc_fut_fn")

    return None


def rootzoneclim(
    dsrun: xr.Dataset,
    ds_obs: xr.Dataset,
    ds_like: xr.Dataset,
    flwdir: xr.DataArray,
    ds_cc_hist: Optional[xr.Dataset] = None,
    ds_cc_fut: Optional[xr.Dataset] = None,
    return_period: list = [2, 3, 5, 10, 15, 20, 25, 50, 60, 100],
    Imax: float = 2.0,
    start_hydro_year: str = "Sep",
    start_field_capacity: str = "Apr",
    LAI: bool = False,
    rootzone_storage: bool = False,
    correct_cc_deficit: bool = False,
    chunksize: int = 100,
    missing_days_threshold: int | None = 330,
):
    """
    Estimates the root zone storage parameter.

    for current observed and (optionally) for future climate-based streamflow data.

    The root zone storage capacity parameter is calculated per subcatchment and is
    converted to a gridded map at model resolution. Optionally, this function
    can return the wflow_sbm parameter vegetation_root_depth by dividing the root zone
    storage parameter by (theta_s - theta_r).

    The method is based on the estimation of maximum annual storage deficits based on
    precipitation and estimated actual evaporation time series, which in turn are
    estimated from observed streamflow data and long-term precipitation and
    potential evap. data, as explained in Bouaziz et al. (2022).
    The main assumption is that vegetation adapts its rootzone storage capacity to
    overcome dry spells with a certain return period
    (typically 20 years for forest ecosystems).
    In response to a changing climtate, it is likely that vegetation also adapts its
    rootzone storage capacity, thereby changing model parameters for future conditions.
    This method also allows to estimate the change in rootzone storage capacity in
    response to a changing climate.

    Parameters
    ----------
    dsrun : xr.Dataset
        Geodataset with streamflow locations and timeseries, named "discharge" (m3/s).
        The geodataset expects the coordinate names "index" (for each station id).
    ds_obs : xr.Dataset
        Dataset with the observed forcing data (precip and pet) [mm/timestep].
    ds_like : xr.Dataset
        Dataset with staticmaps at model resolution.
    flwdir : FlwDirRaster
        flwdir object
    ds_cc_hist : xr.Dataset
        Dataset with the simulated historical forcing data (precip and pet) \
[mm/timestep],
        based on a climate model.
        The default is None.
    ds_cc_fut : xr.Dataset
        Dataset with the simulated future climate forcing data (precip and pet) \
[mm/timestep],
        based on a climate model.
        The default is None.
    return_period : list
        List with one or more values indicating the return period(s) (in
        years) for which the rootzone storage depth should be calculated.
        The default is [2,3,5,10,15,20,25,50,60,100]
    Imax : float
        The maximum interception storage capacity [mm].
        The default is 2 mm.
    start_hydro_year : str
        The start month (abbreviated to the first three letters of the month,
        starting with a capital letter) of the hydrological year.
        The default is "Sep".
    start_field_capacity : str
        The end of the wet season / commencement of dry season. This is the
        moment when the soil is at field capacity, i.e. there is no storage
        deficit yet.
        The default is "Apr".
    rootzone_storage : bool
        Boolean to indicate whether the rootzone storage maps
        should be stored in the staticmaps or not. The default is False.
    LAI : bool
        Determine whether the LAI will be used to determine Imax.
        Requires to have run setup_laimaps.
        The default is False.
    chunksize : int
        Chunksize on time dimension for processing data (not for saving to
        disk!). A default value of 100 is used on the time dimension.
    correct_cc_deficit : bool
        Determines whether a bias-correction of the future deficit should be
        applied. If the climate change scenario and hist period are bias-corrected,
        this should probably be set to False.
    missing_days_threshold: int, optional
            Minimum number of days within a year for that year to be counted in the
            long-term Budyko analysis.

    Returns
    -------
    ds_out : xr.Dataset
        Dataset containing root zone storage capacity (optional) and
        vegetation_root_depth for several forcing and return periods.
    gdf_basins_all : GeoDataFrame
        Geodataframe containing the root zone storage capacity values for
        each basin before filling NaN.


    References
    ----------
    Bouaziz, L. J. E., Aalbers, E. E., Weerts, A. H., Hegnauer, M., Buiteveld,
    H., Lammersen, R., Stam, J., Sprokkereef, E., Savenije, H. H. G. and
    Hrachowitz, M. (2022). Ecosystem adaptation to climate change: the
    sensitivity of hydrological predictions to time-dynamic model parameters,
    Hydrology and Earth System Sciences, 26(5), 1295-1318. DOI:
    10.5194/hess-26-1295-2022.
    """
    # Start with some initial checks
    check_inputs(
        start_hydro_year, start_field_capacity, dsrun, ds_obs, ds_cc_hist, ds_cc_fut
    )

    if LAI == True:
        # Create a new xr dataset containing the interception pars
        intercep_vars = ds_like.LAI.to_dataset(name="LAI")
        intercep_vars["wood_storage"] = ds_like["wood_storage"]
        intercep_vars["leaf_storage"] = ds_like["leaf_storage"]

    # Set the output dataset at model resolution
    ds_out = xr.Dataset(coords=ds_like.raster.coords)
    x_dim = ds_out.raster.x_dim
    y_dim = ds_out.raster.y_dim

    # Make a basin map containing all subcatchments from geodataset of observed
    # streamflow
    x_dim_dsrun = dsrun.vector.x_name
    y_dim_dsrun = dsrun.vector.y_name
    x, y, ids = dsrun[x_dim_dsrun].values, dsrun[y_dim_dsrun].values, dsrun.index.values
    gdf_basins = pd.DataFrame()

    # Loop over basins and get per gauge location a polygon of the upstream
    # area.
    for i, id in enumerate(dsrun.index.values):
        ds_basin_single = flw.basin_map(
            ds_like,
            flwdir,
            ids=ids[i],
            xy=(x[i], y[i]),
            stream=ds_like["rivmsk"],
        )[0]
        ds_basin_single.name = int(dsrun.index.values[i])
        ds_basin_single.raster.set_crs(ds_like.raster.crs)
        gdf_basin_single = ds_basin_single.raster.vectorize()
        gdf_basins = pd.concat([gdf_basins, gdf_basin_single])
    # Set index to catchment id
    gdf_basins.index = gdf_basins.value.astype("int")

    # Add the catchment area to gdf_basins and sort in a descending order
    # make sure to also snap to river when retrieving areas
    idxs_gauges = flwdir.snap(xy=(x, y), mask=ds_like["rivmsk"].values)[0]
    areas_uparea = ds_like["uparea"].values.flat[idxs_gauges]
    df_areas = pd.DataFrame(index=ids, data=areas_uparea * 1e6, columns=["area"])
    gdf_basins = pd.concat([gdf_basins, df_areas], axis=1)
    gdf_basins = gdf_basins.sort_values(by="area", ascending=False)
    # drop basins where area is NaN.
    gdf_basins = gdf_basins[(gdf_basins["area"] >= 0)]

    # calculate mean areal precip and pot evap for the full upstream area of each gauge.
    # loop over ds_obs, ds_cc_hist and ds_cc_fut as they might have
    # different coordinate systems and then merge.
    ds_sub_obs = ds_obs.raster.zonal_stats(gdf_basins, stats=["mean"])
    logger.info("Computing zonal statistics for obs, this can take a while")
    ds_sub_obs = ds_sub_obs.compute()

    if ds_cc_hist is not None:
        ds_sub_cc_hist = ds_cc_hist.raster.zonal_stats(gdf_basins, stats=["mean"])
        logger.info("Computing zonal statistics for cc_hist, this can take a while")
        ds_sub_cc_hist = ds_sub_cc_hist.compute()

    if ds_cc_fut is not None:
        ds_sub_cc_fut = ds_cc_fut.raster.zonal_stats(gdf_basins, stats=["mean"])
        logger.info("Computing zonal statistics for cc_fut, this can take a while")
        ds_sub_cc_fut = ds_sub_cc_fut.compute()

    # Concatenate all forcing types (obs, cc_hist, cc_fut) into a
    # xr dataset after zonal stats
    if ds_cc_hist is not None and ds_cc_fut is not None:
        ds_sub = xr.concat(
            [ds_sub_obs, ds_sub_cc_hist, ds_sub_cc_fut],
            pd.Index(["obs", "cc_hist", "cc_fut"], name="forcing_type"),
        )
    else:
        ds_sub = xr.concat([ds_sub_obs], pd.Index(["obs"], name="forcing_type"))

    # Also get the zonal statistics of the intercep_vars
    if LAI == True:
        intercep_vars_sub = intercep_vars.raster.zonal_stats(gdf_basins, stats=["mean"])
        intercep_vars_sub = intercep_vars_sub.compute()
        # Determine the Imax for every time step in the LAI data
        intercep_vars_sub["Imax"] = (
            intercep_vars_sub["wood_storage_mean"]
            + intercep_vars_sub["LAI_mean"] * intercep_vars_sub["leaf_storage_mean"]
        )
    else:
        intercep_vars_sub = None

    # Get the time step of the datasets and make sure they all have a daily
    # time step. If not, resample.
    time_step = 86400  # in seconds
    if (ds_sub.time[1].values - ds_sub.time[0].values).astype("timedelta64[s]").astype(
        np.int32
    ) != time_step:
        ds_sub = ds_sub.resample(time="1D").sum("time", skipna=True)
    if (dsrun.time[1].values - dsrun.time[0].values).astype("timedelta64[s]").astype(
        np.int32
    ) != time_step:
        dsrun = dsrun.resample(time="1D").mean("time", skipna=True)

    # Determine effective precipitation, interception evporation and canopy
    # storage
    logger.info(
        "Determine effective precipitation, interception evporation and canopy storage"
    )
    ds_sub = determine_Peffective_Interception_explicit(
        ds_sub, Imax=Imax, intercep_vars_sub=intercep_vars_sub
    )

    # Add specific discharge per location-index value to ds_sub
    # First, sort dsrun based on descending subcatchment area (which is already
    # done in ds_sub)
    dsrun = dsrun.sel(index=ds_sub.index.values)
    # Get the specific discharge (mm/timestep) per location in order to have
    # everything in mm/timestep
    dsrun = dsrun.assign(
        specific_Q=dsrun["discharge"].transpose("time", "index")
        / np.array(gdf_basins["area"])
        * time_step
        * 1000.0
    )
    # Add specific discharge to ds_sub
    ds_sub = xr.merge([ds_sub, dsrun["specific_Q"].to_dataset()], compat="override")

    # Rechunk data
    ds_sub = ds_sub.chunk(
        chunks={"index": len(ds_sub.index), "forcing_type": 1, "time": chunksize}
    )

    # Get year sums of ds_sub
    # a threshold is used to use only years with sufficient data
    ds_sub_annual = ds_sub.resample(time=f"AS-{start_hydro_year}").sum(
        "time", skipna=True, min_count=missing_days_threshold
    )

    # Determine discharge coefficient, the aridity index and the evaporative
    # index
    ds_sub_annual = determine_budyko_curve_terms(
        ds_sub_annual,
    )
    # set runoff coefficient of cc_hist equal to runoff coeff of obs
    if correct_cc_deficit == True:
        ds_sub_annual["discharge_coeff"].loc[dict(forcing_type="cc_hist")] = (
            ds_sub_annual["discharge_coeff"].sel(forcing_type="obs")
        )

    # Determine omega
    logger.info("Calculating the omega values, this can take a while")
    ds_sub_annual = determine_omega(ds_sub_annual)

    # Determine future discharge ratio for cc_fut if ds_cc_fut exists
    if ds_cc_fut is not None:
        ds_sub_annual = fut_discharge_coeff(ds_sub_annual, correct_cc_deficit)

    # Determine long-term interception, potential evaporation and tranpsiration;
    # use runoff coefficient instead of Qobs to calculate actual evaporation for
    # each climate projection
    transpiration_long_term = (
        ds_sub_annual["precip_effective"].mean("time")
        - ds_sub_annual["precip_mean"].mean("time") * ds_sub_annual["discharge_coeff"]
    )  # ds_annual['Qobs_mm'].mean('time')
    interception_long_term = ds_sub_annual["evap_interception"].mean("time")
    pet_long_term = ds_sub_annual["pet_mean"].mean("time")

    # Determine the transpiration on the finer time step (e.g. daily)
    ds_sub["transpiration"] = (
        (ds_sub["pet_mean"] - ds_sub["evap_interception"])
        * transpiration_long_term
        / (pet_long_term - interception_long_term)
    )
    transpiration_long_term = None
    interception_long_term = None
    pet_long_term = None

    # Determine storage deficit
    logger.info("Determining the storage deficit")
    ds_sub = determine_storage_deficit(ds_sub, correct_cc_deficit)

    # From the storage deficit, determine the rootzone storage capacity using
    # a Gumbel distribution.
    logger.info("Calculating the Gumbel distribution and rootzone storage capacity")
    # Determine the yearly minima in storage deficit (and make sure the values
    # are positive)
    storage_deficit_annual = -(
        ds_sub["storage_deficit"]
        .resample(time=f"AS-{start_field_capacity}")
        .min("time", skipna=True)
    )

    # A counter will be used to only use years with sufficient days containing
    # data for the Gumbel distribution
    storage_deficit_count = (
        ds_sub["storage_deficit"]
        .resample(time=f"AS-{start_field_capacity}")
        .count("time")
    )

    # Subsequently, determine the Gumbel distribution
    gumbel = gumbel_su_calc_xr(
        storage_deficit_annual,
        storage_deficit_count.sel(time=storage_deficit_count.time[1:]),
        return_period=return_period,
        threshold=missing_days_threshold,
    )

    # Create a new geopandas dataframe, which will be used to rasterize the
    # results
    ds_basins_all = flw.basin_map(
        ds_like,
        flwdir,
        ids=ids,
        xy=(x, y),
        stream=ds_like["rivmsk"],
    )[0]
    ds_basins_all.raster.set_crs(ds_like.raster.crs)
    gdf_basins_all = ds_basins_all.raster.vectorize()
    gdf_basins_all.index = gdf_basins_all.value.astype("int")

    # Add the area and sort by area (from large to small)
    # use previous df
    gdf_basins_all = pd.concat([gdf_basins_all, df_areas], axis=1)
    gdf_basins_all = gdf_basins_all.sort_values(by="area", ascending=False)
    # again drop basins where area is NaN
    gdf_basins_all = gdf_basins_all[(gdf_basins_all["area"] >= 0)]

    # Add the rootzone storage to gdf_basins_all, per forcing type and return
    # period
    for return_period in gumbel.RP.values:
        for forcing_type in gumbel.forcing_type.values:
            gdf_basins_all[f"rootzone_storage_{forcing_type}_{str(return_period)}"] = (
                gumbel["rootzone_storage"].sel(
                    RP=return_period, forcing_type=forcing_type
                )
            )
            # Make sure to give the NaNs a value, otherwise they will become 0.0
            gdf_basins_all[f"rootzone_storage_{forcing_type}_{str(return_period)}"] = (
                gdf_basins_all[
                    f"rootzone_storage_{forcing_type}_{str(return_period)}"
                ].fillna(-999)
            )

    # Rasterize this (from large subcatchments to small ones)
    for return_period in gumbel.RP.values:
        for forcing_type in gumbel.forcing_type.values:
            da_area = ds_like.raster.rasterize(
                gdf=gdf_basins_all,
                col_name=f"rootzone_storage_{forcing_type}_{str(return_period)}",
                nodata=-999,
                all_touched=True,
            ).to_dataset(name="rasterized_temp")
            # Fill up the not a numbers with the data from a downstream point
            out_raster = pyflwdir.FlwdirRaster.fillnodata(
                flwdir,
                data=da_area["rasterized_temp"],
                nodata=-999,
                direction="up",
                how="max",
            )
            # Make sure to fill up full domain with value of most downstream point
            # that contains values
            fill_value = None
            for value in gdf_basins_all[
                f"rootzone_storage_{forcing_type}_{str(return_period)}"
            ]:
                if value > 0.0:
                    if fill_value is None:
                        fill_value = value
            out_raster = np.where(out_raster == -999.0, fill_value, out_raster)
            # Store the rootzone_storage in ds_out is rootzone_storage flag is
            # set to True.
            if rootzone_storage == True:
                ds_out[f"meta_rootzone_storage_{forcing_type}_{str(return_period)}"] = (
                    (y_dim, x_dim),
                    out_raster,
                )

            denom = ds_like["theta_s"].values - ds_like["theta_r"].values
            eps = 1e-6  # numerical safety threshold

            # Safe division: NaN where denominator is ~0
            with np.errstate(divide="ignore", invalid="ignore"):
                vrdepth = np.divide(
                    out_raster,
                    denom,
                    out=np.full_like(out_raster, np.nan, dtype=float),
                    where=np.abs(denom) > eps,
                )
            # Store the vegetation_root_depth in ds_out
            ds_out[f"vegetation_root_depth_{forcing_type}_{str(return_period)}"] = (
                (y_dim, x_dim),
                vrdepth,
            )

    return ds_out, gdf_basins_all



"""Workflows to connect a wflow model to a 1D model."""

import logging

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from hydromt.gis import flw
from pyflwdir import FlwdirRaster
from scipy.ndimage import binary_erosion
from shapely.geometry import LineString, MultiLineString, Point
from shapely.ops import snap

from hydromt_wflow.utils import planar_operation_in_utm

logger = logging.getLogger(f"hydromt.{__name__}")


__all__ = ["wflow_1dmodel_connection"]


def flwdir_mask_to_subbasins(
    flwdir_mask: FlwdirRaster,
    area_max: float,
    ds_model: xr.Dataset,
) -> xr.DataArray:
    """
    Split the common basin between wflow model and 1D model into subbasins.

    The basin is split into subbasins with an area equivalent to area_max [km2].
    """
    logger.info(
        f"Deriving lateral subbasins based on subbasin area threshold: {area_max} km2"
    )
    # calculate subbasins based on area_max threshold
    subbas, _ = flwdir_mask.subbasins_area(area_max)
    da_subbasins = xr.DataArray(
        data=subbas.astype(np.int32),
        dims=ds_model.raster.dims,
        coords=ds_model.raster.coords,
    )
    da_subbasins.raster.set_nodata(0)
    da_subbasins.raster.set_crs(ds_model.raster.crs)
    return da_subbasins


def derive_riv1d_edges(
    riv1d: gpd.GeoDataFrame, ds_model: xr.Dataset, flwdir: FlwdirRaster, **kwargs
) -> gpd.GeoDataFrame:
    """
    Derive the edges of riv1d and snap them to the wflow river.

    First derive the edges of the riv1d (start and end points of each linestring).
    Then snap them to the closest downstream wflow river cell using the
    :py:meth:`hydromt.gis.flw.gauge_map` method.
    Finally, clean up the edges that were snapped to the main river (eg tributary
    starting to close to the main river) by keeping only the most downstream edge
    and the most upstream one.

    Note: this function restricts the use to one river 1d and 1 wflow subcatchment as
    we only consider one downstream edge/edge point for 1d river.
    """
    # get the edges of the riv1d
    riv1d_edges = riv1d.geometry.apply(lambda x: Point(x.coords[0]))
    riv1d_edges = pd.concat(
        [riv1d_edges, riv1d.geometry.apply(lambda x: Point(x.coords[-1]))]
    )
    # find geometry that are unique in riv1d_edges
    riv1d_edges = gpd.GeoDataFrame(
        riv1d_edges[~riv1d_edges.duplicated(keep=False)],
        crs=riv1d.crs,
        geometry="geometry",
    )

    _, idxs, ids = flw.gauge_map(
        ds_model,
        xy=(riv1d_edges.geometry.x, riv1d_edges.geometry.y),
        stream=ds_model["rivmsk"].values,
        flwdir=flwdir,
        **kwargs,
    )
    points = gpd.points_from_xy(*ds_model.raster.idx_to_xy(idxs))
    riv1d_edges = gpd.GeoDataFrame(
        index=ids.astype(np.int32), geometry=points, crs=ds_model.raster.crs
    )

    # Some of the river edges may have been snapped to the main river
    # (eg tributary starting to close to the main river).
    riv1d_edges["uparea"] = ds_model["uparea"].raster.sample(riv1d_edges).values
    # Sort by upstream area descending
    riv1d_edges = riv1d_edges.sort_values(by="uparea", ascending=False)
    # Find the ID of the most downstream edge
    riv1d_edges_outlet_id = riv1d_edges.index[0]
    # Rasterize the riv1d_edges
    riv1d_edges_raster = ds_model.raster.rasterize(riv1d_edges, col_name="index")
    mask = flwdir.downstream(riv1d_edges_raster.values)
    # Find the downstream ID of each edge
    idxs_edge, _ = flwdir.snap(
        xy=(riv1d_edges.geometry.x, riv1d_edges.geometry.y),
        mask=mask,
    )
    edges_on_main_river = mask.flat[idxs_edge]

    # Keep only the edges that are not on the main river and the most downstream edge
    riv1d_edges = riv1d_edges[
        ~riv1d_edges.index.isin(edges_on_main_river)
        | (riv1d_edges.index == riv1d_edges_outlet_id)
    ]

    return riv1d_edges


def connect_subbasin_area(
    ds_model: xr.Dataset,
    riv1d: gpd.GeoDataFrame,
    flwdir: FlwdirRaster,
    area_max: float,
    include_river_boundaries: bool,
    **kwargs,
) -> tuple[FlwdirRaster, gpd.GeoDataFrame]:
    """
    Connect 1d river to Wflow by creating subbasins based on area_max threshold.

    First derive the river edges of the 1d river and snap them to the wflow river.
    Then derive the subbasins corresponding to the river edges.
    Then filter which subbasins are the upstream ones (tributaries) and which ones
    are the downstream ones (main river) and should be split into subbasins.

    The common basin between wflow and the 1d river is then split into subbasins of
    area_max size. Tributaries within the common basin that are larger than area_max are
    derived separately and will results into a tributary inflow point rather than a
    lateral inflow subbasin.

    This function results into a geodataframe of tributaries including the ones upstream
    of the common basin if include_river_boundaries is True and the ones within the
    common basin that have an area larger than area_max. The second output is the rest
    of the common basin where the tributaries have been masked out.
    """
    logger.info("Linking 1D river to wflow river")

    # 1. Derive the river edges / boundaries and snap to the wflow river
    riv1d_edges = derive_riv1d_edges(riv1d, ds_model, flwdir, **kwargs)

    # 2. Derive the subbasins corresponding to the river edges
    da_edges_subbas, _ = flw.basin_map(
        ds_model,
        flwdir=flwdir,
        xy=(riv1d_edges.geometry.x, riv1d_edges.geometry.y),
        ids=riv1d_edges.index,
    )
    da_edges_subbas.raster.set_crs(ds_model.raster.crs)
    # convert to gdf
    gdf_edges_subbas = da_edges_subbas.raster.vectorize()

    # 3. Filter which subbasins are the upstream ones (tributaries)
    # and which ones are the downstream ones (main river)
    # and should be split into subbasins
    # First intersect riv1d with gdf_edges_subbas
    rivmerge = (
        gpd.overlay(riv1d, gdf_edges_subbas)
        .explode(index_parts=True)
        .reset_index(drop=True)
    )
    # Compute len of river
    if rivmerge.crs.is_geographic:
        rivmerge["len"] = rivmerge.geometry.to_crs(3857).length
    else:
        rivmerge["len"] = rivmerge.geometry.length
    # Groupby value and sum length
    rivmerge = rivmerge.groupby("value")["len"].sum()
    # Select the subcatch where rivlength is more than 5times rivlen_avg
    riv_mask = ds_model["rivmsk"].values == 1
    rivlen_avg = ds_model["rivlen"].values[riv_mask].mean()
    subids = rivmerge.index[rivmerge > rivlen_avg * 5].values
    subcatch_to_split = gdf_edges_subbas[gdf_edges_subbas["value"].isin(subids)]
    subcatch_to_split = subcatch_to_split.to_crs(ds_model.raster.crs)
    da_subcatch_to_split = ds_model.raster.rasterize(
        subcatch_to_split, col_name="value"
    ).astype(np.float32)

    # First tributaries are the edges that are not included in the subcatch_to_split
    gdf_tributaries = riv1d_edges[~riv1d_edges.index.isin(subids)]

    # 4. Derive a mask of gdf_riv in the subcatch_to_split wflow rivers
    # compute flw_path to mask out the river included in dfm 1d network
    xy = (
        gdf_tributaries.geometry.x.values.tolist(),
        gdf_tributaries.geometry.y.values.tolist(),
    )

    # Get paths
    flowpaths, _ = flwdir.path(xy=xy, max_length=None, unit="m", direction="down")
    feats = flwdir.geofeatures(flowpaths)
    gdf_paths = gpd.GeoDataFrame.from_features(feats, crs=ds_model.raster.crs)
    gdf_paths.index = np.arange(1, len(gdf_paths) + 1)
    # Create a mask with subcatch id in subcatch_to_split and flowpath is nodata
    da_flwpaths = ds_model.raster.rasterize(gdf_paths)
    da_flwpaths = da_flwpaths.where(
        da_subcatch_to_split != da_subcatch_to_split.raster.nodata,
        da_flwpaths.raster.nodata,
    )

    # 5. Derive the tributaries
    # Find tributaries
    logger.info("Deriving tributaries")
    trib_msk = da_subcatch_to_split.where(
        da_flwpaths == da_flwpaths.raster.nodata, da_subcatch_to_split.raster.nodata
    )
    trib_msk = trib_msk.where(
        (trib_msk != trib_msk.raster.nodata)
        & (ds_model["uparea"] > area_max)
        & (flwdir.downstream(da_flwpaths) != da_flwpaths.raster.nodata),
        trib_msk.raster.nodata,
    )
    # Make sure we vectorize to single cells and not to polygons for adjacent cells
    trib_msk = trib_msk.stack(z=(trib_msk.raster.y_dim, trib_msk.raster.x_dim))
    nodata = trib_msk.raster.nodata
    trib_msk = trib_msk.where(trib_msk != nodata, drop=True)
    trib_msk.values = np.arange(1, len(trib_msk) + 1)
    trib_msk = trib_msk.unstack(fill_value=nodata)
    trib_msk = trib_msk.reindex_like(da_subcatch_to_split, fill_value=nodata).astype(
        np.int32
    )

    gdf_trib = trib_msk.raster.vectorize()
    # Test if there gdf_trib is empty
    if gdf_trib.empty:
        logger.info("No tributaries found")
        if not include_river_boundaries:
            gdf_tributaries = gpd.GeoDataFrame()
    else:
        gdf_trib = gdf_trib.to_crs(ds_model.raster.crs)
        centroid = planar_operation_in_utm(gdf_trib, lambda geom: geom.centroid)
        gdf_trib["geometry"] = centroid

        # Merge with gdf_tributary if include_river_boundaries
        # else only keep intersecting tributaries
        if include_river_boundaries:
            gdf_tributaries = pd.concat(
                [gdf_tributaries, gdf_trib.drop(["value"], axis=1)]
            )
        else:
            gdf_tributaries = gdf_trib.drop(["value"], axis=1)
        gdf_tributaries.index = np.arange(1, len(gdf_tributaries) + 1)

    # 6. Mask the tributaries out of the subatch_to_split map
    if not gdf_tributaries.empty:
        # Derive the tributary basin map
        da_trib_subbas, _ = flw.basin_map(
            ds_model,
            flwdir=flwdir,
            xy=(gdf_tributaries.geometry.x, gdf_tributaries.geometry.y),
        )
        da_trib_subbas.raster.set_crs(ds_model.raster.crs)
        # Mask tributaries
        da_flwdir_mask = ds_model["flwdir"].where(
            (da_subcatch_to_split != da_subcatch_to_split.raster.nodata)
            & (da_trib_subbas == da_trib_subbas.raster.nodata),
            ds_model["flwdir"].raster.nodata,
        )
    else:
        # Mask subcatch to split only
        da_flwdir_mask = ds_model["flwdir"].where(
            da_subcatch_to_split != da_subcatch_to_split.raster.nodata,
            ds_model["flwdir"].raster.nodata,
        )
    flwdir_mask = flw.flwdir_from_da(da_flwdir_mask)

    return (flwdir_mask, gdf_tributaries)


def connect_nodes(
    ds_model: xr.Dataset,
    gdf_riv: gpd.GeoDataFrame,
    flwdir: xr.DataArray,
    flwdir_mask: xr.DataArray,
    **kwargs,
) -> xr.DataArray:
    """Derive wflow subbasins for each 1d river node."""
    # Get the nodes from gdf_riv
    logger.info("Deriving subbasins based on 1D river nodes snapped to wflow river")
    # from multiline to line
    gdf_riv = gdf_riv.explode(ignore_index=True, index_parts=False)
    nodes = []
    for bi, branch in gdf_riv.iterrows():
        nodes.append([Point(branch.geometry.coords[0]), bi])  # start
        nodes.append([Point(branch.geometry.coords[-1]), bi])  # end
    gdf_nodes = gpd.GeoDataFrame(
        nodes, columns=["geometry", "river_id"], crs=gdf_riv.crs
    )
    # Drop duplicates geometry
    gdf_nodes = gdf_nodes[~gdf_nodes.geometry.duplicated(keep="first")]
    gdf_nodes.index = np.arange(1, len(gdf_nodes) + 1)
    # Snap the nodes to the wflow river
    _, idxs, ids = flw.gauge_map(
        ds_model,
        xy=(gdf_nodes.geometry.x, gdf_nodes.geometry.y),
        stream=ds_model["rivmsk"].values,
        flwdir=flwdir,
        **kwargs,
    )
    # Derive subbasins
    da_subbasins, _ = flw.basin_map(
        ds_model,
        flwdir=flwdir_mask,
        idxs=idxs,
        ids=ids,
        stream=ds_model["rivmsk"].values,
    )
    da_subbasins.raster.set_crs(ds_model.raster.crs)
    return da_subbasins


def buffer_basin_mask(
    basin_mask: xr.DataArray, basin_buffer_cells: int
) -> xr.DataArray:
    """Buffer the basin mask raster by a specified number of cells."""
    # Retain raster shape using binary erosion
    basin_mask_values = binary_erosion(
        input=basin_mask.values,
        structure=np.ones((3, 3)),
        iterations=basin_buffer_cells,
    ).astype(basin_mask.dtype)

    # Replace values with shrinked mask
    basin_mask = basin_mask.copy()
    basin_mask.values = basin_mask_values
    if np.sum(basin_mask_values) == 0:
        raise ValueError(
            f"Basin mask is empty after buffering with {basin_buffer_cells} cells. "
            "Consider using a smaller value for 'basin_buffer_cells'."
        )
    return basin_mask


def snap_river_endpoints(
    gdf_riv: gpd.GeoDataFrame,
    geom_snapping_tolerance: float,
) -> gpd.GeoDataFrame:
    """Snap river endpoints that are close to each other."""
    logger.info(
        f"Snapping river segments using geom_snapping_tolerance = "
        f"{geom_snapping_tolerance}."
    )

    # Extract endpoints of each line geometry
    endpoints = []
    for geom in gdf_riv.geometry:
        if isinstance(geom, LineString):
            endpoints.append(Point(geom.coords[0]))  # start point
            endpoints.append(Point(geom.coords[-1]))  # end point

    # Create a GeoSeries of endpoints
    endpoint_gs = gpd.GeoSeries(endpoints)

    # Snap each line geometry to nearby endpoints
    snapped_geometries = []
    snapped_count = 0
    for geom in gdf_riv.geometry:
        snapped_geom = geom
        for point in endpoint_gs:
            new_geom = snap(snapped_geom, point, geom_snapping_tolerance)
            if new_geom != snapped_geom:
                snapped_geom = new_geom
                snapped_count += 1
        snapped_geometries.append(snapped_geom)

    # Update the geometry in the GeoDataFrame
    gdf_riv.geometry = snapped_geometries
    logger.info(
        f"{snapped_count} river segments were snapped. "
        "Update `geom_snapping_tolerance` if this leads to issues."
    )

    return gdf_riv


def subbasin_preprocess_river_geometry(
    gdf_riv: gpd.GeoDataFrame,
    basin_buffer_cells: int,
    geom_snapping_tolerance: float,
) -> gpd.GeoDataFrame:
    """Preprocess the river geometry by snapping and merging segments."""
    # 0. Preprocess the river geometry
    # Check for MultiLineStrings
    if (gdf_riv.geom_type == "MultiLineString").any():
        logger.warning(
            "The river geometry contains one or more MultiLineStrings after "
            "clipping by the basin extent and will now be snapped. Consider increasing "
            f"'basin_buffer_cells' (currently {basin_buffer_cells} cells) if issues "
            "arise later in the process."
        )
        logger.info("Connecting rivers after basin clipping.")
        # we turn these into LineStrings again using its points
        gdf_riv["geometry"] = gdf_riv.geometry.apply(
            lambda geom: LineString(
                [coord for line in geom.geoms for coord in line.coords]
            )
            if isinstance(geom, MultiLineString)
            else geom
        )

        assert not (gdf_riv.geom_type == "MultiLineString").any()

    # Snapping
    if geom_snapping_tolerance > 0:
        gdf_riv = snap_river_endpoints(
            gdf_riv=gdf_riv,
            geom_snapping_tolerance=geom_snapping_tolerance,
        )

    if gdf_riv.empty:
        raise ValueError(
            "No river segments remaining in 'gdf_riv' after preprocessing. "
            "Consider using less strict requirements or check 'gdf_riv'."
        )
    return gdf_riv


def wflow_1dmodel_connection(
    gdf_riv: gpd.GeoDataFrame,
    ds_model: xr.Dataset,
    connection_method: str = "subbasin_area",
    area_max: float = 30.0,
    basin_buffer_cells: int = 0,
    geom_snapping_tolerance: float = 0.001,
    add_tributaries: bool = True,
    include_river_boundaries: bool = True,
    **kwargs,
) -> xr.Dataset:
    """
    Connect wflow to a 1D model by deriving linked subcatchs (and tributaries).

    There are two methods to connect models:

    - `subbasin_area`:
        creates subcatchments linked to the 1d river based on an
        area threshold (area_max) for the subbasin size. With this method, if a
        tributary is larger than the area_max, it will be connected to the 1d river
        directly.
    - `nodes`:
        subcatchments are derived based on the 1driver nodes (used as gauges
        locations). With this method, large tributaries can also be derived separately
        using the add_tributaries option and adding a area_max threshold for the
        tributaries.

    If `add_tributary` option is on, you can decide to include or exclude the upstream
    boundary of the 1d river as an additional tributary using the
    `include_river_boundaries` option.

    River edges or river nodes are snapped to the closest downstream wflow river
    cell using the :py:meth:`hydromt.gis.flw.gauge_map` method.

    Parameters
    ----------
    gdf_riv : gpd.GeoDataFrame
        River geometry.
    ds_model : xr.Dataset
        Model dataset with 'flwdir', 'rivmsk', 'rivlen', 'basins' and 'uparea'
        (optional).
    connection_method : str, default subbasin_area
        Method to connect wflow to the 1D model. Available methods are
        {'subbasin_area', 'nodes'}.
    area_max : float, default 10.0
        Maximum area [km2] of the subbasins to connect to the 1D model in km2 with
        connection_method **subbasin_area** or
        **nodes** with add_tributaries set to True.
    basin_buffer_cells : int, default 0
        Number of cells to use when clipping the 1d river geometry to the basin extent.
        This can be used to not include river geometries near the basin border.
    geom_snapping_tolerance : float, default 0.1
        Distance used to determine whether to snap parts of the 1d river geometry that
        are close to each other. This can be useful if some of the tributaries of the
        1D river are not perfectly connected to the main river.
    add_tributaries : bool, default True
        If True, derive tributaries for the subbasins larger than area_max. Always True
        for **subbasin_area** method.
    include_river_boundaries : bool, default True
        If True, include the upstream boundary(ies) of the 1d river as an additional
        tributary(ies).
    **kwargs
        Additional keyword arguments passed to the snapping method
        hydromt.gis.flw.gauge_map. See its documentation for more information.

    Returns
    -------
    ds_out: xr.Dataset
        Dataset with variables 'subcatch' for the subbasin map, 'subcatch_riv' for the
        subbasin map masked with river cells to be able to save river output with wflow
        and 'gauges' for the tributaries outflow locations (add_tributaries True or
        subbasin_area method).

    See Also
    --------
    hydromt.gis.flw.gauge_map
    """
    # Check variables in 'ds_model'
    dvars_model = ["flwdir", "rivmsk", "rivlen", "basins"]
    if not np.all([v in ds_model for v in dvars_model]):
        raise ValueError(f"One or more variables missing from ds_model: {dvars_model}")
    # Derive flwdir
    flwdir = flw.flwdir_from_da(ds_model["flwdir"])
    # Check for uparea and derive on the fly if needed
    if "uparea" not in ds_model:
        logger.info(
            "upstream area map not found in 'ds_model' and will be derived on the fly."
        )
        uparea = flwdir.upstream_area(unit="km2")
        ds_model["uparea"] = xr.Variable(ds_model.raster.dims, uparea)

    # Check that the max_area is larger than the wflow river threshold
    # Get from attrs if available for newer wflow models built with hydromt
    riv_upa = ds_model["rivmsk"].attrs.get("river_upa", None)
    if riv_upa is None:
        # Derive from the uparea and rivmsk
        riv_upa = xr.where(ds_model["rivmsk"] > 0, ds_model["uparea"], np.nan)
        riv_upa = float(riv_upa.min())
    if area_max < riv_upa:
        new_area_max = np.ceil(riv_upa / 0.5) * 0.5
        # Add 5% buffer to avoid getting subbasins slightly smaller than riv_upa
        # and therefore which won't have river cells
        new_area_max = new_area_max * 1.05
        logger.warning(
            f"The area_max {area_max} is smaller than the minimum upstream area of "
            f"the wflow river {riv_upa} which means tributaries will "
            "not be connected to the wflow river. Changing and setting area_max to "
            f"{new_area_max} km2. "
            f"To keep {area_max} km2 threshold, please update the wflow model to "
            "include more detailed rivers."
        )
        area_max = new_area_max

    # Reproject
    if gdf_riv.crs != ds_model.raster.crs:
        gdf_riv = gdf_riv.to_crs(ds_model.raster.crs)
    # merge multilinestrings in gdf_riv to linestrings to ease processing later on
    if any(gdf_riv.geometry.apply(lambda geom: isinstance(geom, MultiLineString))):
        logger.debug(
            "'gdf_riv' contains MultiLineStrings which will be converted into"
            "LineStrings."
        )
        gdf_riv = gdf_riv.explode(index_parts=True).reset_index(drop=True)

    # Basin mask and clip river geometry
    basin_mask = ds_model["basins"]
    if basin_buffer_cells > 0:
        basin_mask = buffer_basin_mask(
            basin_mask=basin_mask, basin_buffer_cells=basin_buffer_cells
        )
    gdf_riv = gdf_riv.clip(basin_mask.raster.vectorize())

    if connection_method == "subbasin_area" or add_tributaries == True:
        riv1d = subbasin_preprocess_river_geometry(
            gdf_riv=gdf_riv,
            basin_buffer_cells=basin_buffer_cells,
            geom_snapping_tolerance=geom_snapping_tolerance,
        )

        # Obtain flwdir_mask from 'riv1d'
        flwdir_mask, gdf_tributaries = connect_subbasin_area(
            ds_model=ds_model,
            riv1d=riv1d,
            flwdir=flwdir,
            area_max=area_max,
            include_river_boundaries=include_river_boundaries,
            **kwargs,
        )
    else:
        # Use the entire model as flwdir_mask
        flwdir_mask, gdf_tributaries = flwdir, gpd.GeoDataFrame()

    # Derive subbasins
    if connection_method == "subbasin_area":
        da_subbasins = flwdir_mask_to_subbasins(
            flwdir_mask=flwdir_mask, area_max=area_max, ds_model=ds_model
        )
    elif connection_method == "nodes":
        da_subbasins = connect_nodes(
            ds_model=ds_model,
            gdf_riv=gdf_riv,
            flwdir=flwdir,
            flwdir_mask=flwdir_mask,
            **kwargs,
        )

    # Check if the output is not covering the entire model domain
    valid_subbasins = (da_subbasins != da_subbasins.raster.nodata).sum()
    valid_basinmask = (ds_model["flwdir"] != ds_model["flwdir"].raster.nodata).sum()

    if valid_subbasins >= valid_basinmask:
        raise ValueError(
            "The entire model domain is assigned to subbasins. If this should not be "
            "the case, try setting other parameter values or modifying 'gdf_riv'."
        )

    # Check if all subbasins have a wflow river cell
    gdf_subcatch = da_subbasins.raster.vectorize()
    has_river = ds_model["rivmsk"].raster.zonal_stats(gdf_subcatch, stats="count")
    # Count how many subcatchments have no river
    no_river = np.sum(has_river["rivmsk_count"].values == 0)
    if no_river > 0:
        raise ValueError(
            f"{no_river} subbasin(s) do not contain a wflow river cell. "
            "Consider refining the river network in the Wflow model or increasing "
            f"the area_max threshold (currently {area_max} km2)."
        )

    da_subbasins.name = "subcatch"
    ds_out = da_subbasins.to_dataset()

    # Subcatchment map for river cells only (to be able to save river outputs in wflow)
    ds_out["subcatch_riv"] = da_subbasins.where(
        ds_model["rivmsk"] > 0, da_subbasins.raster.nodata
    )

    # Add tributaries
    if not gdf_tributaries.empty:
        ds_out["gauges"] = ds_out.raster.rasterize(
            gdf_tributaries, col_name="index", nodata=0
        )

    return ds_out



"""Landuse workflows for Wflow plugin."""

import logging
import os
from typing import Optional

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from shapely.geometry import box

from hydromt_wflow.workflows.demand import create_grid_from_bbox

logger = logging.getLogger(f"hydromt.{__name__}")


__all__ = [
    "landuse",
    "landuse_from_vector",
    "lai",
    "create_lulc_lai_mapping_table",
    "lai_from_lulc_mapping",
    "add_paddy_to_landuse",
    "add_planted_forest_to_landuse",
    "validate_lulc_vars",
    "LULC_VARS_MAPPING",
]


_RESAMPLING = {
    "landuse": "mode",
    "lai": "average",
    "vegetation_feddes_alpha_h1": "mode",
}

LULC_VARS_MAPPING = {
    "landuse": None,
    "vegetation_kext": "vegetation_canopy__light_extinction_coefficient",
    "land_manning_n": "land_surface_water_flow__manning_n_parameter",
    "soil_compacted_fraction": "compacted_soil__area_fraction",
    "vegetation_root_depth": "vegetation_root__depth",
    "vegetation_leaf_storage": "vegetation__specific_leaf_storage",
    "vegetation_wood_storage": "vegetation_wood_water__storage_capacity",
    "land_water_fraction": "land_water_covered__area_fraction",
    "vegetation_crop_factor": "vegetation__crop_factor",
    "vegetation_feddes_alpha_h1": "vegetation_root__feddes_critical_pressure_head_h1_reduction_coefficient",  # noqa: E501
    "vegetation_feddes_h1": "vegetation_root__feddes_critical_pressure_head_h1",
    "vegetation_feddes_h2": "vegetation_root__feddes_critical_pressure_head_h2",
    "vegetation_feddes_h3_high": "vegetation_root__feddes_critical_pressure_head_h3_high",  # noqa: E501
    "vegetation_feddes_h3_low": "vegetation_root__feddes_critical_pressure_head_h3_low",
    "vegetation_feddes_h4": "vegetation_root__feddes_critical_pressure_head_h4",
    "erosion_usle_c": "soil_erosion__usle_c_factor",
}


def validate_lulc_vars(lulc_vars: list[str]):
    """Throw an error if any lulc_vars are incorrect.

    Parameters
    ----------
    lulc_vars : list[str]
        The lulc_vars that are given to functions like setup_landuse.

    Raises
    ------
        ValueError: An error that indicates which values are not allowed for lulc_vars.
    """
    invalid_vars = set(lulc_vars).difference(LULC_VARS_MAPPING.keys())
    if invalid_vars:
        raise ValueError(
            f"Invalid lulc_vars: {invalid_vars}. "
            f"Allowed values are: {list(LULC_VARS_MAPPING.keys())}."
        )


def landuse(
    da: xr.DataArray,
    ds_like: xr.Dataset,
    df: pd.DataFrame,
    params: Optional[list] = None,
):
    """Return landuse map and related parameter maps.

    The parameter maps are prepared based on landuse map and
    mapping table as provided in the generic data folder of hydromt.

    Parameters
    ----------
    da : xarray.DataArray
        DataArray containing LULC classes.
    ds_like : xarray.DataArray
        Dataset at model resolution.

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing gridded landuse based maps
    """
    keys = df.index.values
    if params is None:
        params = [p for p in df.columns if p != "description"]
    elif not np.all(np.isin(params, df.columns)):
        missing = [p for p in params if p not in df.columns]
        raise ValueError(f"Parameter(s) missing in mapping file: {missing}")
    # setup ds out
    ds_out = xr.Dataset(coords=ds_like.raster.coords)

    # setup reclass method
    def reclass(x):
        return np.vectorize(d.get)(x, nodata)

    da = da.raster.interpolate_na(method="nearest")
    # apply for each parameter
    for param in params:
        method = _RESAMPLING.get(param, "average")
        values = df[param].values
        nodata = values[-1]  # NOTE values is set in last row
        d = dict(zip(keys, values))  # NOTE global param in reclass method
        logger.info(f"Deriving {param} using {method} resampling (nodata={nodata}).")
        da_param = xr.apply_ufunc(
            reclass, da, dask="parallelized", output_dtypes=[values.dtype]
        )
        da_param.attrs.update(_FillValue=nodata)  # first set new nodata values
        ds_out[param] = da_param.raster.reproject_like(
            ds_like, method=method
        )  # then resample

    return ds_out


def landuse_from_vector(
    gdf: gpd.GeoDataFrame,
    ds_like: xr.Dataset,
    df: pd.DataFrame,
    params: Optional[list] = None,
    lulc_res: float | int | None = None,
    all_touched: bool = False,
    buffer: int = 1000,
    lulc_out: Optional[str] = None,
):
    """
    Derive several wflow maps based on vector landuse-landcover (LULC) data.

    The vector lulc data is first rasterized to a raster map at the model resolution
    or at a higher resolution specified in ``lulc_res``.

    Lookup table `df` columns are converted to lulc classes model
    parameters based on literature. The data is remapped at its rasterized resolution
    and then resampled to the model resolution using the average value, unless noted
    differently.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame containing LULC classes.
    ds_like : xarray.Dataset
        Dataset at model resolution.
    df : pd.DataFrame
        Mapping table with landuse values.
    params : list of str, optional
        List of parameters to derive, by default None
    lulc_res : float or int, optional
        Resolution of the rasterized LULC data, by default None (use model resolution)
    all_touched : bool, optional
        If True, all pixels touched by the polygon will be burned in, by default False
    buffer : int, optional
        Buffer in meters to add around the bounding box of the vector data, by default
        1000.
    lulc_out : str, optional
        Path to save the rasterised original landuse map to file, by default None.

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing gridded landuse based maps
    """
    # intersect with bbox
    bounds = gpd.GeoDataFrame(
        geometry=[box(*ds_like.raster.bounds)], crs=ds_like.raster.crs
    )
    bounds = bounds.to_crs(3857).buffer(buffer).to_crs(gdf.crs)
    gdf = gdf.overlay(gpd.GeoDataFrame(geometry=bounds), how="intersection")

    # rasterize the vector data
    logger.info("Rasterizing landuse map")
    if lulc_res is None:
        gdf_reproj = gdf.to_crs(ds_like.raster.crs)
        grid_like = create_grid_from_bbox(
            gdf_reproj.total_bounds,
            res=max(np.abs(ds_like.raster.res)),
            crs=ds_like.raster.crs,
            align=True,
        )
    else:
        grid_like = create_grid_from_bbox(
            gdf.total_bounds,
            res=lulc_res,
            crs=gdf.crs,
            align=True,
        )
    # get the nodata values of the landuse (last row in the df)
    nodata = df["landuse"].values[-1]
    da = grid_like.raster.rasterize(
        gdf,
        col_name="landuse",
        nodata=nodata,
        all_touched=all_touched,
        dtype="int32",
    )
    if lulc_out is not None:
        logger.info(f"Saving rasterized landuse map to {lulc_out}")
        os.makedirs(os.path.dirname(lulc_out), exist_ok=True)
        da.raster.to_raster(lulc_out)

    # derive the landuse maps
    ds_out = landuse(da, ds_like, df, params=params)

    return ds_out


def lai(da: xr.DataArray, ds_like: xr.Dataset):
    """Return climatology of Leaf Area Index (LAI).

    The following maps are calculated:
    - LAI

    Parameters
    ----------
    da : xarray.DataArray or xarray.Dataset
        LAI array containing LAI values.
    ds_like : xarray.DataArray
        Dataset at model resolution.

    Returns
    -------
    da_out : xarray.DataArray
        Dataset containing resampled LAI maps
    """
    if isinstance(da, xr.Dataset) and "LAI" in da:
        da = da["LAI"]
    elif not isinstance(da, xr.DataArray):
        raise ValueError("lai method requires a DataArray or Dataset with LAI array")
    method = "average"
    nodata = da.raster.nodata
    logger.info(f"Deriving {da.name} using {method} resampling (nodata={nodata}).")
    da = da.astype(np.float32)
    # Assuming missing values correspond to: bare soil, urban and snow (LAI=0.0)
    da = da.where(da.values != nodata).fillna(0.0)
    da_out = da.raster.reproject_like(ds_like, method=method)
    da_out.attrs.update(_FillValue=nodata)
    return da_out


def create_lulc_lai_mapping_table(
    da_lulc: xr.DataArray,
    da_lai: xr.DataArray,
    sampling_method: str = "any",
    lulc_zero_classes: list[int] = [],
) -> pd.DataFrame:
    """
    Derive LAI values per landuse class.

    Parameters
    ----------
    da_lulc : xr.DataArray
        Landuse map.
    da_lai : xr.DataArray
        Cyclic LAI map.
    sampling_method : str, optional
        Resampling method for the LULC data to LAI resolution.
        Two methods are supported:

        * 'any' (default): if any cell of the desired landuse class is present in the
            resampling window (even just one), it will be used to derive LAI values.
            This method is less exact but will provide LAI values for all landuse
            classes for the high resolution landuse map.
        * 'mode': the most frequent value in the resampling window is
            used. This method is less precise as for cells with a lot of different
            landuse classes, the most frequent value might still be only a small
            fraction of the cell. More landuse classes should however be covered and
            it can always be used with the landuse map of the wflow model instead of
            the original high resolution one.
        * 'q3': only cells with the most frequent value (mode) and that cover 75%
            (q3) of the resampling window will be used. This method is more exact but
            for small basins, you may have less or no samples to derive LAI values
            for some classes.
    lulc_zero_classes : list of int, optional
        List of landuse classes that should have zero for leaf area index values
        for example waterbodies, open ocean etc. For very high resolution landuse
        maps, urban surfaces and bare areas can be included here as well.
        By default empty.

    Returns
    -------
    df_lai_mapping : pd.DataFrame
        Mapping table with LAI values per landuse class. One column for each month and
        one line per landuse class. The number of samples used to derive the mapping
        values is also added to a `samples` column in the dataframe.
    """
    # check the method values
    if sampling_method not in ["any", "mode", "q3"]:
        raise ValueError(f"Unsupported resampling method: {sampling_method}")

    # process the lai da
    if "dim0" in da_lai.dims:
        da_lai = da_lai.rename({"dim0": "time"})
    da_lai = da_lai.raster.mask_nodata()
    da_lai = da_lai.fillna(
        0
    )  # use zeros to better represent city and open water surfaces

    # landuse
    da_lulc.name = "landuse"
    lulc_classes = np.unique(da_lulc.values)

    # Initialise the outputs
    df_lai_mapping = None

    if sampling_method != "any":
        # The data can already be resampled to the LAI resolution
        da_lulc_mode = da_lulc.raster.reproject_like(da_lai, method="mode")
        if sampling_method == "q3":
            # Filter mode cells that cover less than 75% of the resampling window
            da_lulc_q3 = da_lulc.raster.reproject_like(da_lai, method="q3")
            da_lulc = da_lulc_mode.where(
                da_lulc_q3 == da_lulc_mode, da_lulc_mode.raster.nodata
            )
        else:
            da_lulc = da_lulc_mode

    # Loop over the landuse classes
    for lulc_id in lulc_classes:
        logger.info(f"Processing landuse class {lulc_id}")

        if lulc_id in lulc_zero_classes:
            logger.info(f"Using zeros for landuse class {lulc_id}")
            df_lai = pd.DataFrame(
                columns=da_lai.time.values,
                data=[[0] * 12],
                index=[lulc_id],
            )
            df_lai.index.name = "landuse"
            n_samples = 0

        else:
            # Select for a specific landuse class
            lu = da_lulc.where(da_lulc == lulc_id, da_lulc.raster.nodata)
            lu = lu.raster.mask_nodata()

            if sampling_method == "any":
                # Resample the landuse data to the LAI resolution
                lu = lu.raster.reproject_like(da_lai, method="mode")

            # Add lai
            lu = lu.to_dataset()
            lu["lai"] = da_lai

            # Stack and remove the nodata values
            lu = lu.stack(z=(lu.raster.y_dim, lu.raster.x_dim)).dropna(
                dim="z", how="all", subset=["landuse"]
            )

            # Count the number of samples
            n_samples = len(lu["z"])
            if n_samples == 0:
                logger.info(
                    f"No samples found for landuse class {lulc_id}. "
                    "Try using a different resampling method."
                )
                df_lai = pd.DataFrame(
                    columns=da_lai.time.values,
                    data=[[0] * 12],
                    index=[lulc_id],
                )
                df_lai.index.name = "landuse"
            else:
                # Compute the mean
                lai_mean_per_lu = np.round(lu["lai"].load().mean(dim="z"), 3)
                # Add the landuse id as an extra dimension
                lai_mean_per_lu = lai_mean_per_lu.expand_dims("landuse")
                lai_mean_per_lu["landuse"] = [lulc_id]

                # Convert to dataframe
                df_lai = lai_mean_per_lu.drop_vars(
                    "spatial_ref", errors="ignore"
                ).to_pandas()

        # Add number of samples in the first column
        df_lai.insert(0, "samples", n_samples)

        # Append to the output
        if df_lai_mapping is None:
            df_lai_mapping = df_lai
        else:
            df_lai_mapping = pd.concat([df_lai_mapping, df_lai])

    return df_lai_mapping


def lai_from_lulc_mapping(
    da: xr.DataArray,
    ds_like: xr.Dataset,
    df: pd.DataFrame,
) -> xr.Dataset:
    """
    Derive LAI values from a landuse map and a mapping table.

    Parameters
    ----------
    da : xr.DataArray
        Landuse map.
    ds_like : xr.Dataset
        Dataset at model resolution.
    df : pd.DataFrame
        Mapping table with LAI values per landuse class. One column for each month and
        one line per landuse class.

    Returns
    -------
    ds_lai : xr.Dataset
        Dataset with LAI values for each month.
    """
    months = np.arange(1, 13)
    df.columns = [int(col) if str(col).isdigit() else col for col in df.columns]
    # Map the monthly LAI values to the landuse map
    ds_lai = landuse(
        da=da,
        ds_like=ds_like,
        df=df,
        params=months,
    )
    # Re-organise the dataset to have a time dimension
    da_lai = ds_lai.to_array(dim="time", name="LAI")

    return da_lai


def add_paddy_to_landuse(
    landuse: xr.DataArray,
    paddy: xr.DataArray,
    paddy_class: int,
    df_mapping: pd.DataFrame,
    df_paddy_mapping: pd.DataFrame,
    output_paddy_class: Optional[int] = None,
) -> tuple[xr.DataArray, pd.DataFrame]:
    """
    Burn paddy fields into landuse map and update mapping table.

    The resulting paddy class in the landuse map will have ID output_paddy_class
    if provided and paddy_class otherwise. The mapping table will be updated with
    the values from the df_paddy_mapping table.

    Parameters
    ----------
    landuse : xr.DataArray
        Landuse map.
    paddy : xr.DataArray
        Paddy fields map.
    paddy_class : int
        ID of the paddy class in the paddy map.
    df_mapping : pd.DataFrame
        Mapping table with landuse values.
    df_paddy_mapping : pd.DataFrame
        Mapping table with paddy values.
    output_paddy_class : int, optional
        ID of the paddy class in the output landuse map. If not provided, the
        paddy_class will be used.

    Returns
    -------
    landuse : xr.DataArray
        Updated landuse map.
    df_mapping : pd.DataFrame
        Updated mapping table.
    """
    # Get output paddy class
    if output_paddy_class is None:
        output_paddy_class = paddy_class

    # Reproject paddy map to landuse resolution
    # if paddy has lower res than landuse, use nearest resampling
    if abs(paddy.raster.res[0]) >= abs(landuse.raster.res[0]):
        paddy = paddy.raster.reproject_like(landuse, method="nearest")
    # else use mode resampling
    else:
        paddy = paddy.raster.reproject_like(landuse, method="mode")

    # Burn in the rice fields in the landuse map
    landuse = landuse.where(paddy != paddy_class, output_paddy_class)

    # Update the mapping table
    df_paddy_mapping.index = [output_paddy_class]
    df_paddy_mapping["landuse"] = output_paddy_class
    # Add the paddy class to the first line of the mapping table
    df_mapping = pd.concat([df_paddy_mapping, df_mapping])

    return landuse, df_mapping


def add_planted_forest_to_landuse(
    planted_forest: gpd.GeoDataFrame,
    ds_like: xr.Dataset,
    planted_forest_c: float = 0.0881,
    orchard_name: str = "Orchard",
    orchard_c: float = 0.2188,
) -> xr.DataArray:
    """
    Update USLE C map with planted forest and orchard data.

    Default USLE C values for planted forest and orchards are derived from Panagos et
    al., 2015 (10.1016/j.landusepol.2015.05.021).
    For harvested forest at different regrowth stages see also Borrelli and Schutt, 2014
    (10.1016/j.geomorph.2013.08.022).

    Parameters
    ----------
    planted_forest : geopandas.GeoDataFrame
        GeoDataFrame containing planted forest data. Required columns are: 'geometry',
        and optionally 'forest_type' to find orchards.
    ds_like : xr.Dataset
        Dataset at model resolution. Required variables are 'usle_c'.
    planted_forest_c : float, optional
        USLE C value for planted forest, by default 0.0881.
    orchard_name : str, optional
        Name of the orchard landuse class, by default "Orchard".
    orchard_c : float, optional
        USLE C value for orchards, by default 0.2188.

    Returns
    -------
    usle_c : xr.DataArray
        Updated USLE C map.
    """
    # Add a usle_c column with default value
    logger.info(
        "Correcting usle_c with planted forest and orchards using {planted_forest_fn}."  # noqa: E501
    )
    planted_forest["usle_c"] = planted_forest_c
    # If forest_type column is available, update usle_c value for orchards
    if "forest_type" in planted_forest.columns:
        planted_forest.loc[planted_forest["forest_type"] == orchard_name, "usle_c"] = (
            orchard_c
        )
    # Rasterize forest data
    usle_c = ds_like.raster.rasterize(
        gdf=planted_forest,
        col_name="usle_c",
        nodata=ds_like["usle_c"].raster.nodata,
        all_touched=False,
    )
    # Cover nodata with the usle_c map from all landuse classes
    usle_c = usle_c.where(
        usle_c != usle_c.raster.nodata,
        ds_like["usle_c"],
    )

    return usle_c



import pytest

import hydromt_wflow.workflows as workflows


def test_validate_lulc_vars_wrong_values():
    with pytest.raises(ValueError, match="Invalid lulc_vars: {'invalid_var'}"):
        workflows.validate_lulc_vars(["landuse", "invalid_var"])


def test_validate_lulc_vars_correct_not_all_values():
    # Should not raise
    workflows.validate_lulc_vars(["landuse", "erosion_usle_c"])


def test_validate_lulc_vars_all_correct_values():
    workflows.validate_lulc_vars(workflows.LULC_VARS_MAPPING.keys())



from pathlib import Path

import pandas as pd

from hydromt_wflow.wflow_sbm import WflowSbmModel


def test_write_and_read_with_custom_staticmaps_folder(tmp_path: Path):
    model = WflowSbmModel(tmp_path)
    model.staticmaps.write("staticmaps/staticmaps.nc")
    test_df = pd.DataFrame({"a": [1, 2, 3]})
    model.tables.set(test_df, name="test")
    model.tables.write()

    assert (tmp_path / "staticmaps" / "test.csv").is_file()

    model2 = WflowSbmModel(tmp_path, mode="r")
    model2.config.set("input.path_static", "staticmaps/staticmaps.nc")
    model2.tables.read()
    pd.testing.assert_frame_equal(model2.tables.data["test"], test_df)


def test_write_and_read_without_staticmaps_folder(tmp_path: Path):
    model = WflowSbmModel(tmp_path)
    test_df = pd.DataFrame({"a": [1, 2, 3]})
    model.tables.set(test_df, name="test")
    model.tables.write()

    assert (tmp_path / "test.csv").is_file()

    model2 = WflowSbmModel(tmp_path, mode="r")
    model2.tables.read()
    pd.testing.assert_frame_equal(model2.tables.data["test"], test_df)
